from logging import Logger

from httpmessageprocessor import HttpMessageProcessorMessageHandler
from httpmessageprocessorutil.automapped import AutoMappedMessageProcessor
from interactor.event import ElectronicPaymentEventInteractor
from interactor.route.electronic_payment.cache import CacheInteractor
from interactor.route.electronic_payment.cache.dto import CacheRequest
from interactor.route.customization.get_styles import GetStylesInteractor
from interactor.route.customization.get_styles.dto import GetStylesRequest
from interactor.route.menu.get_images import GetImagesInteractor
from interactor.route.menu.get_images.dto import GetImagesRequest
from interactor.route.menu.get_navigation import GetNavigationInteractor
from interactor.route.menu.get_navigation.dto import GetNavigationRequest
from interactor.route.menu.get_product_info import GetProductInfoInteractor
from interactor.route.menu.get_product_info.dto import GetProductInfoRequest
from interactor.route.menu.get_products import GetProductsInteractor
from interactor.route.menu.get_products.dto import GetProductsRequest
from interactor.route.menu.get_ruptured_products import GetRupturedProducts
from interactor.route.menu.get_ruptured_products.dto import GetRupturedProductsRequest
from interactor.route.menu.get_banners import GetBannersInteractor
from interactor.route.menu.get_banners.dto import GetBannersRequest
from interactor.route.menu.get_up_sell_part_codes import GetUpSellPartCodesInteractor
from interactor.route.menu.get_up_sell_part_codes.dto import GetUpSellPartCodesRequest
from interactor.route.order.add_items import AddItemsInteractor
from interactor.route.order.add_items.dto import AddItemsRequest
from interactor.route.order.add_properties import AddPropertiesInteractor
from interactor.route.order.add_properties.dto import AddPropertiesRequest
from interactor.route.order.apply_coupon import ApplyCouponInteractor
from interactor.route.order.apply_coupon.dto import ApplyCouponRequest
from interactor.route.order.cancel import OrderCancelInteractor
from interactor.route.order.cancel.dto import OrderCancelRequest
from interactor.route.order.change_items_quantity import ChangeItemsQuantityInteractor
from interactor.route.order.change_items_quantity.dto import ChangeItemsQuantityRequest
from interactor.route.order.close_shopping_cart import CloseShoppingCartInteractor
from interactor.route.order.close_shopping_cart.dto import CloseShoppingCartRequest
from interactor.route.order.create import OrderCreateInteractor
from interactor.route.order.create.dto import OrderCreateRequest
from interactor.route.order.delete_line import OrderDeleteLineInteractor
from interactor.route.order.delete_line.dto import OrderDeleteLineRequest
from interactor.route.order.finalize import OrderFinalizeInteractor
from interactor.route.order.finalize.dto import OrderFinalizeRequest
from interactor.route.order.get_detail import OrderGetDetailInteractor
from interactor.route.order.get_detail.dto import OrderGetDetailRequest
from interactor.route.payment.insert import OrderInsertPaymentInteractor
from interactor.route.payment.insert.dto import OrderInsertPaymentRequest
from interactor.route.order.reopen_shopping_cart import ReopenShoppingCartInteractor
from interactor.route.order.reopen_shopping_cart.dto import ReopenShoppingCartRequest
from interactor.route.order.remove_coupon import RemoveCouponInteractor
from interactor.route.order.remove_coupon.dto import RemoveCouponRequest
from interactor.route.payment.get_payment_types import GetPaymentTypesInteractor
from interactor.route.payment.get_payment_types.dto import GetPaymentTypesRequest
from interactor.route.electronic_payment.read_messages import ReadMessagesInteractor
from interactor.route.electronic_payment.read_messages.dto import ReadMessagesRequest
from interactor.route.electronic_payment.clear_messages import ClearMessagesInteractor
from interactor.route.electronic_payment.clear_messages.dto import ClearMessagesRequest
from interactor.route.electronic_payment.send_responses import SendResponsesInteractor
from interactor.route.electronic_payment.send_responses.dto import SendResponsesRequest
from interactor.route.customization.get_text import GetTextInteractor
from interactor.route.customization.get_text.dto import GetTextRequest
from interactor.route.peripheral.print_text import PrintTextInteractor
from interactor.route.peripheral.print_text.dto import PrintTextRequest
from interactor.route.peripheral.read_scanner import ReadScannerInteractor
from interactor.route.peripheral.read_scanner.dto import ReadScannerRequest
from interactor.route.store.get_status import GetStoreInteractor
from interactor.route.store.get_status.dto import GetStoreRequest
from interactor.route.store.get_config import GetStoreConfigInteractor
from interactor.route.store.get_config.dto import GetStoreConfigRequest
from interactor.route.payment.cancel import CancelPaymentInteractor
from interactor.route.payment.cancel.dto import CancelPaymentRequest
from mbcontextmessagehandler import MbContextMessageBus
from messagebus import Message, Event
from messagehandler import MessageHandlerBuilder, MessageHandler
from messageprocessor import MessageProcessor, DefaultMessageProcessorExecutorFactory, MessageProcessorMessageHandler
from messageprocessorutil import LoggingProcessorCallback
from model.config import Config
from model.event import ListenedEvents
from model.exception import (
    ValidationError,
    AuthValidationErrors,
    ExternalServiceErrors,
    OperationErrors,
    BadRequestError,
)
from model.response import Responses
from processor.event.electronic_payment import (
    ElectronicPaymentCloseDialogProcessor,
    ElectronicPaymentShowQRCodeProcessor,
    ElectronicPaymentShowChoicesProcessor,
    ElectronicPaymentShowInputProcessor,
    ElectronicPaymentShowMessageProcessor,
    ElectronicPaymentShowQuestionProcessor
)
from pyscripts_cache import PyscriptsCache
from repository import PosRepository
from services import (
    FiscalService,
    OrderService,
    ApiService,
    CouponService,
    ImagesService,
    PaymentService,
    ProductInfoService
)
from simplehttprouter import Route
from typing import Dict, Optional


class ServiceMessageHandlerBuilder(MessageHandlerBuilder):

    def __init__(self, logger, config, message_bus, flask_server=False):
        # type: (Logger, Config, MbContextMessageBus, bool) -> None

        self.logger = logger
        self.config = config
        self.message_bus = message_bus
        self.flask_server = flask_server

        self.pos_repository = None
        self.fiscal_service = None
        self.coupon_service = None
        self.order_service = None
        self.api_service = None
        self.modifier_service = None
        self.images_service = None
        self.payment_service = None
        self.product_info_service = None

        self.pyscripts_cache = None

        self.cache_menu_ruptured_items = {"last_request_date": None, "data": None}
        self.cache_menu_products = None
        self.cache_menu_navigation = []
        self.electronic_messages = []

        self.callbacks = None
        self.message_handler = None

    def build_singletons(self):
        # type: () -> None

        self.callbacks = [LoggingProcessorCallback(self.logger)]

        if self.flask_server:
            self.pos_repository = PosRepository(self.message_bus, self.logger, self.config)
            self.fiscal_service = FiscalService(self.message_bus, self.logger)
            self.coupon_service = CouponService(self.message_bus, self.logger)
            self.pyscripts_cache = PyscriptsCache(self.message_bus.mbcontext)
            self.order_service = OrderService()
            self.images_service = ImagesService()
            self.payment_service = PaymentService(
                self.message_bus,
                self.config
            )
            self.product_info_service = ProductInfoService(self.message_bus)

            self.fiscal_service.get_electronic_payments_id()
        else:
            self.api_service = ApiService(self.config, self.logger)

            events_processors = self._build_events_processors()
            default_message_processor_executor_factory = DefaultMessageProcessorExecutorFactory(self.callbacks)
            self.message_handler = MessageProcessorMessageHandler(events_processors,
                                                                  None,
                                                                  default_message_processor_executor_factory)

    def build_message_handler(self, message=None, event=None):
        # type: (Optional[Message], Optional[Event]) -> MessageHandler

        if self.flask_server:
            processors_routes = self._build_processors_routes()
            http_message_processor = HttpMessageProcessorMessageHandler(route_processor=processors_routes,
                                                                        callbacks=self.callbacks)
            return http_message_processor
        else:
            return self.message_handler

    def destroy_message_handler(self, message_handler):
        # type: (MessageHandler) -> None

        pass

    def destroy_singletons(self):
        # type: () -> None

        pass

    def _build_events_processors(self):
        # type: () -> Optional[Dict[str, MessageProcessor]]

        electronic_payment_event_interactor = ElectronicPaymentEventInteractor(self.api_service)

        electronic_payment_close_dialog_processor = \
            ElectronicPaymentCloseDialogProcessor(electronic_payment_event_interactor)
        electronic_payment_show_qrcode_processor = \
            ElectronicPaymentShowQRCodeProcessor(electronic_payment_event_interactor)
        electronic_payment_show_choices_processor = \
            ElectronicPaymentShowChoicesProcessor(electronic_payment_event_interactor)
        electronic_payment_show_input_processor = \
            ElectronicPaymentShowInputProcessor(electronic_payment_event_interactor)
        electronic_payment_show_message_processor = \
            ElectronicPaymentShowMessageProcessor(electronic_payment_event_interactor)
        electronic_payment_show_question_processor = \
            ElectronicPaymentShowQuestionProcessor(electronic_payment_event_interactor)

        events_processors = {
            ListenedEvents.ELECTRONIC_PAYMENT_CLOSE_DIALOGS: electronic_payment_close_dialog_processor,
            ListenedEvents.ELECTRONIC_PAYMENT_SHOW_QRCODE: electronic_payment_show_qrcode_processor,
            ListenedEvents.ELECTRONIC_PAYMENT_SHOW_CHOICES: electronic_payment_show_choices_processor,
            ListenedEvents.ELECTRONIC_PAYMENT_SHOW_INPUT: electronic_payment_show_input_processor,
            ListenedEvents.ELECTRONIC_PAYMENT_SHOW_MESSAGE: electronic_payment_show_message_processor,
            ListenedEvents.ELECTRONIC_PAYMENT_SHOW_QUESTION: electronic_payment_show_question_processor
        }

        return events_processors

    def _build_processors_routes(self):
        # type: () -> Dict[Route, MessageProcessor]

        customization_get_styles_interactor = GetStylesInteractor(
            logger=self.logger,
            config=self.config
        )
        order_add_items_interactor = \
            AddItemsInteractor(self.logger, self.config, self.pos_repository)
        order_cancel_interactor = \
            OrderCancelInteractor(self.logger, self.config, self.pos_repository)
        order_delete_line_interactor = \
            OrderDeleteLineInteractor(self.logger, self.config, self.pos_repository)
        order_close_shopping_cart_interactor = \
            CloseShoppingCartInteractor(self.logger, self.config, self.pos_repository, self.order_service)
        order_create_interactor = \
            OrderCreateInteractor(self.logger, self.config, self.pos_repository)
        order_add_properties_interactor = \
            AddPropertiesInteractor(self.logger, self.config, self.pos_repository)
        order_finalize_interactor = \
            OrderFinalizeInteractor(self.logger, self.config, self.pos_repository, self.fiscal_service)
        order_get_detail_interactor = \
            OrderGetDetailInteractor(self.logger, self.config, self.pos_repository, self.order_service)
        order_insert_payment_interactor = OrderInsertPaymentInteractor(
            self.logger,
            self.config,
            self.pos_repository,
            self.fiscal_service,
            self.payment_service
        )
        order_change_items_quantity_interactor = \
            ChangeItemsQuantityInteractor(self.logger, self.config, self.pos_repository)
        order_reopen_shopping_cart_interactor = ReopenShoppingCartInteractor(
            logger=self.logger,
            config=self.config,
            pos_repository=self.pos_repository,
            order_service=self.order_service
        )
        order_apply_coupon_interactor = ApplyCouponInteractor(
            logger=self.logger,
            config=self.config,
            pos_repository=self.pos_repository,
            coupon_service=self.coupon_service,
            order_service=self.order_service
        )
        order_remove_coupon_interactor = RemoveCouponInteractor(
            logger=self.logger,
            config=self.config,
            pos_repository=self.pos_repository,
            order_service=self.order_service
        )
        menu_get_products_interactor = \
            GetProductsInteractor(self.logger, self.config, self.pos_repository, self.cache_menu_products)
        menu_get_ruptured_products = \
            GetRupturedProducts(self.logger, self.config, self.pos_repository, self.cache_menu_ruptured_items)
        menu_get_navigation_interactor = GetNavigationInteractor(
            logger=self.logger,
            config=self.config,
            pos_repository=self.pos_repository,
            cache_navigation=self.cache_menu_navigation
        )
        menu_get_images_interactor = GetImagesInteractor(self.logger, self.config, self.images_service)
        menu_get_banners_interactor = GetBannersInteractor(
            logger=self.logger,
            config=self.config
        )
        menu_get_up_sell_part_codes_interactor = GetUpSellPartCodesInteractor(
            logger=self.logger,
            config=self.config
        )
        menu_get_product_info_interactor = GetProductInfoInteractor(
            logger=self.logger,
            config=self.config,
            pos_repository=self.pos_repository,
            product_info_service=self.product_info_service
        )
        payment_get_payment_types_interactor = GetPaymentTypesInteractor(
            self.logger,
            self.config,
            self.pos_repository,
            self.payment_service
        )
        peripheral_electronic_payment_read_messages_interactor = \
            ReadMessagesInteractor(
                self.logger,
                self.config,
                self.pos_repository,
                self.fiscal_service,
                self.electronic_messages
            )
        peripheral_electronic_payment_clear_messages_interactor = \
            ClearMessagesInteractor(
                self.logger,
                self.config,
                self.pos_repository,
                self.electronic_messages
            )
        peripheral_electronic_payment_send_responses_interactor = \
            SendResponsesInteractor(
                self.logger,
                self.config,
                self.pos_repository,
                self.fiscal_service,
                self.electronic_messages
            )
        peripheral_print_text_interactor = \
            PrintTextInteractor(self.logger, self.config, self.pos_repository)
        peripheral_read_scanner_interactor = \
            ReadScannerInteractor(self.logger, self.config, self.pos_repository)
        status_get_store_interactor = \
            GetStoreInteractor(self.logger, self.config, self.pos_repository)
        cache_peripheral_electronic_payment_interactor = \
            CacheInteractor(self.logger, self.config, self.electronic_messages)

        customization_get_store_config_interactor = GetStoreConfigInteractor(
            logger=self.logger,
            config=self.config,
            pos_repository=self.pos_repository
        )

        customization_get_text_interactor = GetTextInteractor(
            self.logger,
            self.config,
            self.pos_repository
        )

        order_cancel_payment_interactor = CancelPaymentInteractor(
            self.logger,
            self.config,
            self.pos_repository
        )

        processor_routes = {
            Route("GET", "/order/getDetail/{posId}"): AutoMappedMessageProcessor(
                OrderGetDetailRequest,
                order_get_detail_interactor,
                {
                    ValidationError: Responses.ValidationError,
                    AuthValidationErrors: Responses.AuthValidation,
                    ExternalServiceErrors: Responses.ExternalServicesError,
                    OperationErrors: Responses.OperationErrors
                }
            ),

            Route("POST", "/order/create/{posId}"): AutoMappedMessageProcessor(
                OrderCreateRequest,
                order_create_interactor,
                {
                    ValidationError: Responses.ValidationError,
                    AuthValidationErrors: Responses.AuthValidation,
                    ExternalServiceErrors: Responses.ExternalServicesError,
                    OperationErrors: Responses.OperationErrors
                },
                success_status=201
            ),

            Route("POST", "/order/addProperties/{posId}"): AutoMappedMessageProcessor(
                AddPropertiesRequest,
                order_add_properties_interactor,
                {
                    ValidationError: Responses.ValidationError,
                    AuthValidationErrors: Responses.AuthValidation,
                    ExternalServiceErrors: Responses.ExternalServicesError,
                    OperationErrors: Responses.OperationErrors
                },
                success_status=204
            ),

            Route("POST", "/order/addItems/{posId}"): AutoMappedMessageProcessor(
                AddItemsRequest,
                order_add_items_interactor,
                {
                    ValidationError: Responses.ValidationError,
                    AuthValidationErrors: Responses.AuthValidation,
                    ExternalServiceErrors: Responses.ExternalServicesError,
                    OperationErrors: Responses.OperationErrors
                },
                success_status=204
            ),

            Route("POST", "/order/changeItemsQuantity/{posId}"): AutoMappedMessageProcessor(
                ChangeItemsQuantityRequest,
                order_change_items_quantity_interactor,
                {
                    ValidationError: Responses.ValidationError,
                    AuthValidationErrors: Responses.AuthValidation,
                    ExternalServiceErrors: Responses.ExternalServicesError,
                    OperationErrors: Responses.OperationErrors
                },
                success_status=204
            ),

            Route("POST", "/order/insertPayment/{posId}"): AutoMappedMessageProcessor(
                OrderInsertPaymentRequest,
                order_insert_payment_interactor,
                {
                    ValidationError: Responses.ValidationError,
                    AuthValidationErrors: Responses.AuthValidation,
                    ExternalServiceErrors: Responses.ExternalServicesError,
                    OperationErrors: Responses.OperationErrors,
                    BadRequestError: Responses.BadRequestError,
                }
            ),

            Route("POST", "/order/registerPayment/{posId}"): AutoMappedMessageProcessor(
                OrderInsertPaymentRequest,
                order_insert_payment_interactor,
                {
                    ValidationError: Responses.ValidationError,
                    AuthValidationErrors: Responses.AuthValidation,
                    ExternalServiceErrors: Responses.ExternalServicesError,
                    OperationErrors: Responses.OperationErrors,
                    BadRequestError: Responses.BadRequestError,
                }),

            Route("DELETE", "/order/cancelPayment/{posId}/{tenderId}"): AutoMappedMessageProcessor(
                CancelPaymentRequest,
                order_cancel_payment_interactor,
                {
                    ValidationError: Responses.ValidationError,
                    AuthValidationErrors: Responses.AuthValidation,
                    ExternalServiceErrors: Responses.ExternalServicesError,
                    OperationErrors: Responses.OperationErrors
                }
            ),

            Route("POST", "/order/cancel/{posId}"): AutoMappedMessageProcessor(
                OrderCancelRequest,
                order_cancel_interactor,
                {
                    ValidationError: Responses.ValidationError,
                    AuthValidationErrors: Responses.AuthValidation,
                    ExternalServiceErrors: Responses.ExternalServicesError,
                    OperationErrors: Responses.OperationErrors
                },
                success_status=204
            ),

            Route("DELETE", "/order/deleteLine/{posId}"): AutoMappedMessageProcessor(
                OrderDeleteLineRequest,
                order_delete_line_interactor,
                {
                    ValidationError: Responses.ValidationError,
                    AuthValidationErrors: Responses.AuthValidation,
                    ExternalServiceErrors: Responses.ExternalServicesError,
                    OperationErrors: Responses.OperationErrors
                },
                success_status=204
            ),

            Route("POST", "/order/closeShoppingCart/{posId}"): AutoMappedMessageProcessor(
                CloseShoppingCartRequest,
                order_close_shopping_cart_interactor,
                {
                    ValidationError: Responses.ValidationError,
                    AuthValidationErrors: Responses.AuthValidation,
                    ExternalServiceErrors: Responses.ExternalServicesError,
                    OperationErrors: Responses.OperationErrors
                }
            ),

            Route("POST", "/order/reopenShoppingCart/{posId}"): AutoMappedMessageProcessor(
                request_type=ReopenShoppingCartRequest,
                interactor=order_reopen_shopping_cart_interactor,
                exception_map={
                    ValidationError: Responses.ValidationError,
                    AuthValidationErrors: Responses.AuthValidation,
                    ExternalServiceErrors: Responses.ExternalServicesError,
                    OperationErrors: Responses.OperationErrors
                }
            ),

            Route("POST", "/order/finalize/{posId}"): AutoMappedMessageProcessor(
                OrderFinalizeRequest,
                order_finalize_interactor,
                {
                    ValidationError: Responses.ValidationError,
                    AuthValidationErrors: Responses.AuthValidation,
                    ExternalServiceErrors: Responses.ExternalServicesError,
                    OperationErrors: Responses.OperationErrors
                },
                success_status=204
            ),

            Route("POST", "/order/applyCoupon/{posId}"): AutoMappedMessageProcessor(
                ApplyCouponRequest,
                order_apply_coupon_interactor,
                {
                    ValidationError: Responses.ValidationError,
                    AuthValidationErrors: Responses.AuthValidation,
                    ExternalServiceErrors: Responses.ExternalServicesError,
                    OperationErrors: Responses.OperationErrors
                },
                success_status=200
            ),

            Route("DELETE", "/order/removeCoupon/{posId}"): AutoMappedMessageProcessor(
                RemoveCouponRequest,
                order_remove_coupon_interactor,
                {
                    ValidationError: Responses.ValidationError,
                    AuthValidationErrors: Responses.AuthValidation,
                    ExternalServiceErrors: Responses.ExternalServicesError,
                    OperationErrors: Responses.OperationErrors
                }
            ),

            Route("GET", "/menu/products"): AutoMappedMessageProcessor(
                GetProductsRequest,
                menu_get_products_interactor,
                {
                    ValidationError: Responses.ValidationError,
                    AuthValidationErrors: Responses.AuthValidation,
                    ExternalServiceErrors: Responses.ExternalServicesError,
                    OperationErrors: Responses.OperationErrors
                },
                response_content_type="application/gzip"
            ),

            Route("GET", "/menu/rupturedProducts"): AutoMappedMessageProcessor(
                GetRupturedProductsRequest,
                menu_get_ruptured_products,
                {
                    ValidationError: Responses.ValidationError,
                    AuthValidationErrors: Responses.AuthValidation,
                    ExternalServiceErrors: Responses.ExternalServicesError,
                    OperationErrors: Responses.OperationErrors
                }
            ),

            Route("GET", "/menu/navigation/{posId}"): AutoMappedMessageProcessor(
                GetNavigationRequest,
                menu_get_navigation_interactor,
                {
                    ValidationError: Responses.ValidationError,
                    AuthValidationErrors: Responses.AuthValidation,
                    ExternalServiceErrors: Responses.ExternalServicesError,
                    OperationErrors: Responses.OperationErrors
                }
            ),

            Route("GET", "/menu/productInfo/{language}"): AutoMappedMessageProcessor(
                GetProductInfoRequest,
                menu_get_product_info_interactor,
                {
                    ValidationError: Responses.ValidationError,
                    AuthValidationErrors: Responses.AuthValidation,
                    ExternalServiceErrors: Responses.ExternalServicesError,
                    OperationErrors: Responses.OperationErrors
                }),

            Route("GET", "/menu/images"): AutoMappedMessageProcessor(
                GetImagesRequest,
                menu_get_images_interactor,
                {
                    ValidationError: Responses.ValidationError,
                    AuthValidationErrors: Responses.AuthValidation,
                    ExternalServiceErrors: Responses.ExternalServicesError,
                    OperationErrors: Responses.OperationErrors
                }
            ),

            Route("GET", "/menu/banners"): AutoMappedMessageProcessor(
                GetBannersRequest,
                menu_get_banners_interactor,
                {
                    AuthValidationErrors: Responses.AuthValidation
                }
            ),

            Route("GET", "/menu/upSellPartCodes"): AutoMappedMessageProcessor(
                GetUpSellPartCodesRequest,
                menu_get_up_sell_part_codes_interactor,
                {
                    AuthValidationErrors: Responses.AuthValidation
                }),

            Route("GET", "/payment/paymentTypes"): AutoMappedMessageProcessor(
                GetPaymentTypesRequest,
                payment_get_payment_types_interactor,
                {
                    ValidationError: Responses.ValidationError,
                    AuthValidationErrors: Responses.AuthValidation,
                    ExternalServiceErrors: Responses.ExternalServicesError,
                    OperationErrors: Responses.OperationErrors
                }
            ),

            Route("POST", "/peripheral/electronicPayment/readMessages/{posId}"): AutoMappedMessageProcessor(
                ReadMessagesRequest,
                peripheral_electronic_payment_read_messages_interactor,
                {
                    ValidationError: Responses.ValidationError,
                    AuthValidationErrors: Responses.AuthValidation,
                    ExternalServiceErrors: Responses.ExternalServicesError,
                    OperationErrors: Responses.OperationErrors
                }
            ),

            Route("POST", "/peripheral/electronicPayment/clearMessages/{posId}"): AutoMappedMessageProcessor(
                ClearMessagesRequest,
                peripheral_electronic_payment_clear_messages_interactor,
                {
                    ValidationError: Responses.ValidationError,
                    AuthValidationErrors: Responses.AuthValidation,
                    ExternalServiceErrors: Responses.ExternalServicesError,
                    OperationErrors: Responses.OperationErrors
                }
            ),

            Route("POST", "/peripheral/electronicPayment/sendResponse/{posId}"): AutoMappedMessageProcessor(
                SendResponsesRequest,
                peripheral_electronic_payment_send_responses_interactor,
                {
                    ValidationError: Responses.ValidationError,
                    AuthValidationErrors: Responses.AuthValidation,
                    ExternalServiceErrors: Responses.ExternalServicesError,
                    OperationErrors: Responses.OperationErrors
                },
                success_status=204
            ),

            Route("POST", "/peripheral/printText/{posId}"): AutoMappedMessageProcessor(
                PrintTextRequest,
                peripheral_print_text_interactor,
                {
                    ValidationError: Responses.ValidationError,
                    AuthValidationErrors: Responses.AuthValidation,
                    ExternalServiceErrors: Responses.ExternalServicesError
                },
                success_status=204
            ),

            Route("GET", "/peripheral/readScanner/{posId}"): AutoMappedMessageProcessor(
                ReadScannerRequest,
                peripheral_read_scanner_interactor,
                {
                    ValidationError: Responses.ValidationError,
                    AuthValidationErrors: Responses.AuthValidation,
                    ExternalServiceErrors: Responses.ExternalServicesError,
                    OperationErrors: Responses.OperationErrors
                },
                success_status=200
            ),

            Route("GET", "/status/store/{posId}"): AutoMappedMessageProcessor(
                GetStoreRequest,
                status_get_store_interactor,
                {
                    ValidationError: Responses.ValidationError,
                    AuthValidationErrors: Responses.AuthValidation,
                    ExternalServiceErrors: Responses.ExternalServicesError,
                    OperationErrors: Responses.OperationErrors
                }
            ),

            Route("POST", "/cache/peripheral/electronicPayment"): AutoMappedMessageProcessor(
                CacheRequest,
                cache_peripheral_electronic_payment_interactor,
                {},
                success_status=204
            ),

            Route("GET", "/customization/getStyles"): AutoMappedMessageProcessor(
                request_type=GetStylesRequest,
                interactor=customization_get_styles_interactor,
                exception_map={
                    ValidationError: Responses.ValidationError,
                    AuthValidationErrors: Responses.AuthValidation,
                    ExternalServiceErrors: Responses.ExternalServicesError,
                    OperationErrors: Responses.OperationErrors
                }
            ),

            Route("GET", "/customization/getText"): AutoMappedMessageProcessor(
                GetTextRequest,
                customization_get_text_interactor,
                {
                    ValidationError: Responses.ValidationError,
                    AuthValidationErrors: Responses.AuthValidation,
                    ExternalServiceErrors: Responses.ExternalServicesError
                }
            ),

            Route("GET", "/customization/storeConfig/{posId}"): AutoMappedMessageProcessor(
                GetStoreConfigRequest,
                customization_get_store_config_interactor,
                {
                    ValidationError: Responses.ValidationError,
                    AuthValidationErrors: Responses.AuthValidation,
                    ExternalServiceErrors: Responses.ExternalServicesError,
                    OperationErrors: Responses.OperationErrors
                }
            )
        }

        return processor_routes



from _NfceXmlPartBuilder import NfceXmlPartBuilder
from _ContextKeys import ContextKeys
from modelcommon.model import Store
from pos_model import Order
from typing import Any, Dict


class TranspBuilder(NfceXmlPartBuilder):
    def __init__(
            self,
            store_model     # type: Store
    ):
        #  type: (...) -> None

        self.store_model = store_model
        street_name = self.store_model.address_street
        street_number = self.store_model.address_number
        neighborhood = self.store_model.neighborhood
        self.address = "{0}, {1} - {2}".format(street_name, street_number, neighborhood)
        self.address = self.address[:60]

    def build_xml(self, order, context):
        # type: (Order, Dict[unicode, Any]) -> unicode

        delivery_partner = context[ContextKeys.delivery_partner]
        mod_freight = 9 if delivery_partner is None else 4
        shipping = ""
        if delivery_partner is not None:
            shipping = self._get_shipping()

        return "<transp><modFrete>{0}</modFrete>{1}</transp>".format(mod_freight, shipping)

    def _get_shipping(
            self
    ):
        # type: (...)-> unicode

        tag = "<transporta>{}</transporta>"
        shipping = "<CNPJ>{}</CNPJ>".format(self.store_model.cnpj)
        shipping += "<xNome>{}</xNome>".format(self.store_model.name[:60])
        shipping += "<IE>{}</IE>".format(self.store_model.state_registration)
        shipping += "<xEnder>{}</xEnder>".format(self.address[:60])
        shipping += "<xMun>{}</xMun>".format(self.store_model.city[:60])
        shipping += "<UF>{}</UF>".format(self.store_model.state)

        return tag.format(shipping)



data = (
'byum',    # 0x00
'byub',    # 0x01
'byubs',    # 0x02
'byus',    # 0x03
'byuss',    # 0x04
'byung',    # 0x05
'byuj',    # 0x06
'byuc',    # 0x07
'byuk',    # 0x08
'byut',    # 0x09
'byup',    # 0x0a
'byuh',    # 0x0b
'beu',    # 0x0c
'beug',    # 0x0d
'beugg',    # 0x0e
'beugs',    # 0x0f
'beun',    # 0x10
'beunj',    # 0x11
'beunh',    # 0x12
'beud',    # 0x13
'beul',    # 0x14
'beulg',    # 0x15
'beulm',    # 0x16
'beulb',    # 0x17
'beuls',    # 0x18
'beult',    # 0x19
'beulp',    # 0x1a
'beulh',    # 0x1b
'beum',    # 0x1c
'beub',    # 0x1d
'beubs',    # 0x1e
'beus',    # 0x1f
'beuss',    # 0x20
'beung',    # 0x21
'beuj',    # 0x22
'beuc',    # 0x23
'beuk',    # 0x24
'beut',    # 0x25
'beup',    # 0x26
'beuh',    # 0x27
'byi',    # 0x28
'byig',    # 0x29
'byigg',    # 0x2a
'byigs',    # 0x2b
'byin',    # 0x2c
'byinj',    # 0x2d
'byinh',    # 0x2e
'byid',    # 0x2f
'byil',    # 0x30
'byilg',    # 0x31
'byilm',    # 0x32
'byilb',    # 0x33
'byils',    # 0x34
'byilt',    # 0x35
'byilp',    # 0x36
'byilh',    # 0x37
'byim',    # 0x38
'byib',    # 0x39
'byibs',    # 0x3a
'byis',    # 0x3b
'byiss',    # 0x3c
'bying',    # 0x3d
'byij',    # 0x3e
'byic',    # 0x3f
'byik',    # 0x40
'byit',    # 0x41
'byip',    # 0x42
'byih',    # 0x43
'bi',    # 0x44
'big',    # 0x45
'bigg',    # 0x46
'bigs',    # 0x47
'bin',    # 0x48
'binj',    # 0x49
'binh',    # 0x4a
'bid',    # 0x4b
'bil',    # 0x4c
'bilg',    # 0x4d
'bilm',    # 0x4e
'bilb',    # 0x4f
'bils',    # 0x50
'bilt',    # 0x51
'bilp',    # 0x52
'bilh',    # 0x53
'bim',    # 0x54
'bib',    # 0x55
'bibs',    # 0x56
'bis',    # 0x57
'biss',    # 0x58
'bing',    # 0x59
'bij',    # 0x5a
'bic',    # 0x5b
'bik',    # 0x5c
'bit',    # 0x5d
'bip',    # 0x5e
'bih',    # 0x5f
'bba',    # 0x60
'bbag',    # 0x61
'bbagg',    # 0x62
'bbags',    # 0x63
'bban',    # 0x64
'bbanj',    # 0x65
'bbanh',    # 0x66
'bbad',    # 0x67
'bbal',    # 0x68
'bbalg',    # 0x69
'bbalm',    # 0x6a
'bbalb',    # 0x6b
'bbals',    # 0x6c
'bbalt',    # 0x6d
'bbalp',    # 0x6e
'bbalh',    # 0x6f
'bbam',    # 0x70
'bbab',    # 0x71
'bbabs',    # 0x72
'bbas',    # 0x73
'bbass',    # 0x74
'bbang',    # 0x75
'bbaj',    # 0x76
'bbac',    # 0x77
'bbak',    # 0x78
'bbat',    # 0x79
'bbap',    # 0x7a
'bbah',    # 0x7b
'bbae',    # 0x7c
'bbaeg',    # 0x7d
'bbaegg',    # 0x7e
'bbaegs',    # 0x7f
'bbaen',    # 0x80
'bbaenj',    # 0x81
'bbaenh',    # 0x82
'bbaed',    # 0x83
'bbael',    # 0x84
'bbaelg',    # 0x85
'bbaelm',    # 0x86
'bbaelb',    # 0x87
'bbaels',    # 0x88
'bbaelt',    # 0x89
'bbaelp',    # 0x8a
'bbaelh',    # 0x8b
'bbaem',    # 0x8c
'bbaeb',    # 0x8d
'bbaebs',    # 0x8e
'bbaes',    # 0x8f
'bbaess',    # 0x90
'bbaeng',    # 0x91
'bbaej',    # 0x92
'bbaec',    # 0x93
'bbaek',    # 0x94
'bbaet',    # 0x95
'bbaep',    # 0x96
'bbaeh',    # 0x97
'bbya',    # 0x98
'bbyag',    # 0x99
'bbyagg',    # 0x9a
'bbyags',    # 0x9b
'bbyan',    # 0x9c
'bbyanj',    # 0x9d
'bbyanh',    # 0x9e
'bbyad',    # 0x9f
'bbyal',    # 0xa0
'bbyalg',    # 0xa1
'bbyalm',    # 0xa2
'bbyalb',    # 0xa3
'bbyals',    # 0xa4
'bbyalt',    # 0xa5
'bbyalp',    # 0xa6
'bbyalh',    # 0xa7
'bbyam',    # 0xa8
'bbyab',    # 0xa9
'bbyabs',    # 0xaa
'bbyas',    # 0xab
'bbyass',    # 0xac
'bbyang',    # 0xad
'bbyaj',    # 0xae
'bbyac',    # 0xaf
'bbyak',    # 0xb0
'bbyat',    # 0xb1
'bbyap',    # 0xb2
'bbyah',    # 0xb3
'bbyae',    # 0xb4
'bbyaeg',    # 0xb5
'bbyaegg',    # 0xb6
'bbyaegs',    # 0xb7
'bbyaen',    # 0xb8
'bbyaenj',    # 0xb9
'bbyaenh',    # 0xba
'bbyaed',    # 0xbb
'bbyael',    # 0xbc
'bbyaelg',    # 0xbd
'bbyaelm',    # 0xbe
'bbyaelb',    # 0xbf
'bbyaels',    # 0xc0
'bbyaelt',    # 0xc1
'bbyaelp',    # 0xc2
'bbyaelh',    # 0xc3
'bbyaem',    # 0xc4
'bbyaeb',    # 0xc5
'bbyaebs',    # 0xc6
'bbyaes',    # 0xc7
'bbyaess',    # 0xc8
'bbyaeng',    # 0xc9
'bbyaej',    # 0xca
'bbyaec',    # 0xcb
'bbyaek',    # 0xcc
'bbyaet',    # 0xcd
'bbyaep',    # 0xce
'bbyaeh',    # 0xcf
'bbeo',    # 0xd0
'bbeog',    # 0xd1
'bbeogg',    # 0xd2
'bbeogs',    # 0xd3
'bbeon',    # 0xd4
'bbeonj',    # 0xd5
'bbeonh',    # 0xd6
'bbeod',    # 0xd7
'bbeol',    # 0xd8
'bbeolg',    # 0xd9
'bbeolm',    # 0xda
'bbeolb',    # 0xdb
'bbeols',    # 0xdc
'bbeolt',    # 0xdd
'bbeolp',    # 0xde
'bbeolh',    # 0xdf
'bbeom',    # 0xe0
'bbeob',    # 0xe1
'bbeobs',    # 0xe2
'bbeos',    # 0xe3
'bbeoss',    # 0xe4
'bbeong',    # 0xe5
'bbeoj',    # 0xe6
'bbeoc',    # 0xe7
'bbeok',    # 0xe8
'bbeot',    # 0xe9
'bbeop',    # 0xea
'bbeoh',    # 0xeb
'bbe',    # 0xec
'bbeg',    # 0xed
'bbegg',    # 0xee
'bbegs',    # 0xef
'bben',    # 0xf0
'bbenj',    # 0xf1
'bbenh',    # 0xf2
'bbed',    # 0xf3
'bbel',    # 0xf4
'bbelg',    # 0xf5
'bbelm',    # 0xf6
'bbelb',    # 0xf7
'bbels',    # 0xf8
'bbelt',    # 0xf9
'bbelp',    # 0xfa
'bbelh',    # 0xfb
'bbem',    # 0xfc
'bbeb',    # 0xfd
'bbebs',    # 0xfe
'bbes',    # 0xff
)



from abc import ABCMeta
from datetime import datetime

from dateutil import tz
from msgbus import MBEasyContext
from persistence import Connection, Driver, DbdException
from typing import Callable, Any, List


class BaseRepository(object):
    __metaclass__ = ABCMeta

    def __init__(self, mb_context):
        # type: (MBEasyContext) -> None
        self.mb_context = mb_context

    def execute_with_connection(self, method, db_name=None, service_name=None):
        # type: (Callable[[Connection], Any], str, str) -> Any

        conn = None
        try:
            db_name = str(db_name) if db_name is not None else None
            conn = Driver().open(self.mb_context, dbname=db_name, service_name=service_name)
            return method(conn)
        finally:
            if conn is not None:
                conn.close()

    def execute_with_transaction(self, method, db_name=None, service_name=None):
        # type: (Callable[[Connection], Any], str, str) -> Any
        def transaction_method(conn):
            # type: (Connection) -> Any

            if conn.get_number_of_instances() != 1 and db_name is None:
                raise DbdException("Cannot open transaction without DB instance")

            conn.transaction_start()
            try:
                conn.query("BEGIN TRANSACTION")
                ret = method(conn)
                conn.query("COMMIT")
                return ret
            except:
                conn.query("ROLLBACK")
                raise
            finally:
                conn.transaction_end()

        return self.execute_with_connection(transaction_method, db_name=db_name, service_name=service_name)

    def execute_in_all_databases(self, method, pos_list):
        # type: (Callable[[Connection], Any], List[int]) -> Any
        ret = []
        for pos in pos_list:
            conn = None
            try:
                conn = Driver().open(self.mb_context, dbname=str(pos))
                ret.append(method(conn))
            finally:
                if conn is not None:
                    conn.close()

        return ret

    def execute_in_all_databases_returning_flat_list(self, method, pos_list):
        # type: (Callable[[Connection], Any], List[int]) -> Any
        inner_ret = self.execute_in_all_databases(method, pos_list)

        ret = []
        for inner_list in inner_ret:
            ret.extend(inner_list)

        return ret


def ensure_iterable(value):
    if isinstance(value, basestring):
        return [value]

    try:
        _ = (e for e in value)
        return value
    except TypeError:
        return [value]


def ensure_list(value):
    if isinstance(value, basestring):
        return [value]

    if isinstance(value, list):
        return value

    try:
        _ = (e for e in value)
        return list(value)
    except TypeError:
        return [value]


def convert_to_dict(iterable):
    ret = {}
    for item in iterable:
        ret[item] = item
    return ret


def convert_from_localtime_to_utc(localtime):
    # type: (datetime) -> datetime
    from_zone = tz.tzlocal()
    to_zone = tz.tzutc()
    localtime = localtime.replace(tzinfo=from_zone)
    utc = localtime.astimezone(to_zone)

    return utc



import json
import os

from logging import Logger

from interactor.route.customization.get_styles.dto import GetStylesRequest
from model.config import Config
from model.customization_styles import CustomizationStyle
from validation_util import validate_api_key


class GetStylesInteractor(object):
    def __init__(
        self,
        logger,     # type: Logger
        config      # type: Config
    ):
        # type: (...) -> None

        self.logger = logger
        self.config = config

    def execute(
        self,
        request     # type: GetStylesRequest
    ):
        # type: (...) -> CustomizationStyle

        validate_api_key(
            request=request,
            secure_api_key_list=self.config.api_key_mapping
        )

        if not self.config.totem_config_path or not os.path.exists(self.config.totem_config_path):
            return CustomizationStyle({})

        config_file = open(self.config.totem_config_path, 'rb')
        data = json.load(config_file)
        style = data.get("style", {})

        return CustomizationStyle(style)



class GetStoreResponse(object):
    def __init__(
            self,
            status,     # type: bool
            menu_id     # type: str
    ):
        # type: (...) -> None

        self.isOnline = status
        self.menuId = menu_id



# -*- coding: utf-8 -*-
"""
This module offers timezone implementations subclassing the abstract
:py:`datetime.tzinfo` type. There are classes to handle tzfile format files
(usually are in :file:`/etc/localtime`, :file:`/usr/share/zoneinfo`, etc), TZ
environment string (in all known formats), given ranges (with help from
relative deltas), local machine timezone, fixed offset timezone, and UTC
timezone.
"""
import datetime
import struct
import time
import sys
import os
import bisect

from six import string_types
from ._common import tzname_in_python2, _tzinfo, _total_seconds
from ._common import tzrangebase, enfold
from ._common import _validate_fromutc_inputs

try:
    from .win import tzwin, tzwinlocal
except ImportError:
    tzwin = tzwinlocal = None

ZERO = datetime.timedelta(0)
EPOCH = datetime.datetime.utcfromtimestamp(0)
EPOCHORDINAL = EPOCH.toordinal()


class tzutc(datetime.tzinfo):
    """
    This is a tzinfo object that represents the UTC time zone.
    """
    def utcoffset(self, dt):
        return ZERO

    def dst(self, dt):
        return ZERO

    @tzname_in_python2
    def tzname(self, dt):
        return "UTC"

    def is_ambiguous(self, dt):
        """
        Whether or not the "wall time" of a given datetime is ambiguous in this
        zone.

        :param dt:
            A :py:class:`datetime.datetime`, naive or time zone aware.


        :return:
            Returns ``True`` if ambiguous, ``False`` otherwise.

        .. versionadded:: 2.6.0
        """
        return False

    @_validate_fromutc_inputs
    def fromutc(self, dt):
        """
        Fast track version of fromutc() returns the original ``dt`` object for
        any valid :py:class:`datetime.datetime` object.
        """
        return dt

    def __eq__(self, other):
        if not isinstance(other, (tzutc, tzoffset)):
            return NotImplemented

        return (isinstance(other, tzutc) or
                (isinstance(other, tzoffset) and other._offset == ZERO))

    __hash__ = None

    def __ne__(self, other):
        return not (self == other)

    def __repr__(self):
        return "%s()" % self.__class__.__name__

    __reduce__ = object.__reduce__


class tzoffset(datetime.tzinfo):
    """
    A simple class for representing a fixed offset from UTC.

    :param name:
        The timezone name, to be returned when ``tzname()`` is called.

    :param offset:
        The time zone offset in seconds, or (since version 2.6.0, represented
        as a :py:class:`datetime.timedelta` object.
    """
    def __init__(self, name, offset):
        self._name = name

        try:
            # Allow a timedelta
            offset = _total_seconds(offset)
        except (TypeError, AttributeError):
            pass
        self._offset = datetime.timedelta(seconds=offset)

    def utcoffset(self, dt):
        return self._offset

    def dst(self, dt):
        return ZERO

    @tzname_in_python2
    def tzname(self, dt):
        return self._name

    @_validate_fromutc_inputs
    def fromutc(self, dt):
        return dt + self._offset

    def is_ambiguous(self, dt):
        """
        Whether or not the "wall time" of a given datetime is ambiguous in this
        zone.

        :param dt:
            A :py:class:`datetime.datetime`, naive or time zone aware.


        :return:
            Returns ``True`` if ambiguous, ``False`` otherwise.

        .. versionadded:: 2.6.0
        """
        return False

    def __eq__(self, other):
        if not isinstance(other, tzoffset):
            return NotImplemented

        return self._offset == other._offset

    __hash__ = None

    def __ne__(self, other):
        return not (self == other)

    def __repr__(self):
        return "%s(%s, %s)" % (self.__class__.__name__,
                               repr(self._name),
                               int(_total_seconds(self._offset)))

    __reduce__ = object.__reduce__


class tzlocal(_tzinfo):
    """
    A :class:`tzinfo` subclass built around the ``time`` timezone functions.
    """
    def __init__(self):
        super(tzlocal, self).__init__()

        self._std_offset = datetime.timedelta(seconds=-time.timezone)
        if time.daylight:
            self._dst_offset = datetime.timedelta(seconds=-time.altzone)
        else:
            self._dst_offset = self._std_offset

        self._dst_saved = self._dst_offset - self._std_offset
        self._hasdst = bool(self._dst_saved)

    def utcoffset(self, dt):
        if dt is None and self._hasdst:
            return None

        if self._isdst(dt):
            return self._dst_offset
        else:
            return self._std_offset

    def dst(self, dt):
        if dt is None and self._hasdst:
            return None

        if self._isdst(dt):
            return self._dst_offset - self._std_offset
        else:
            return ZERO

    @tzname_in_python2
    def tzname(self, dt):
        return time.tzname[self._isdst(dt)]

    def is_ambiguous(self, dt):
        """
        Whether or not the "wall time" of a given datetime is ambiguous in this
        zone.

        :param dt:
            A :py:class:`datetime.datetime`, naive or time zone aware.


        :return:
            Returns ``True`` if ambiguous, ``False`` otherwise.

        .. versionadded:: 2.6.0
        """
        naive_dst = self._naive_is_dst(dt)
        return (not naive_dst and
                (naive_dst != self._naive_is_dst(dt - self._dst_saved)))

    def _naive_is_dst(self, dt):
        timestamp = _datetime_to_timestamp(dt)
        return time.localtime(timestamp + time.timezone).tm_isdst

    def _isdst(self, dt, fold_naive=True):
        # We can't use mktime here. It is unstable when deciding if
        # the hour near to a change is DST or not.
        #
        # timestamp = time.mktime((dt.year, dt.month, dt.day, dt.hour,
        #                         dt.minute, dt.second, dt.weekday(), 0, -1))
        # return time.localtime(timestamp).tm_isdst
        #
        # The code above yields the following result:
        #
        # >>> import tz, datetime
        # >>> t = tz.tzlocal()
        # >>> datetime.datetime(2003,2,15,23,tzinfo=t).tzname()
        # 'BRDT'
        # >>> datetime.datetime(2003,2,16,0,tzinfo=t).tzname()
        # 'BRST'
        # >>> datetime.datetime(2003,2,15,23,tzinfo=t).tzname()
        # 'BRST'
        # >>> datetime.datetime(2003,2,15,22,tzinfo=t).tzname()
        # 'BRDT'
        # >>> datetime.datetime(2003,2,15,23,tzinfo=t).tzname()
        # 'BRDT'
        #
        # Here is a more stable implementation:
        #
        if not self._hasdst:
            return False

        # Check for ambiguous times:
        dstval = self._naive_is_dst(dt)
        fold = getattr(dt, 'fold', None)

        if self.is_ambiguous(dt):
            if fold is not None:
                return not self._fold(dt)
            else:
                return True

        return dstval

    def __eq__(self, other):
        if not isinstance(other, tzlocal):
            return NotImplemented

        return (self._std_offset == other._std_offset and
                self._dst_offset == other._dst_offset)

    __hash__ = None

    def __ne__(self, other):
        return not (self == other)

    def __repr__(self):
        return "%s()" % self.__class__.__name__

    __reduce__ = object.__reduce__


class _ttinfo(object):
    __slots__ = ["offset", "delta", "isdst", "abbr",
                 "isstd", "isgmt", "dstoffset"]

    def __init__(self):
        for attr in self.__slots__:
            setattr(self, attr, None)

    def __repr__(self):
        l = []
        for attr in self.__slots__:
            value = getattr(self, attr)
            if value is not None:
                l.append("%s=%s" % (attr, repr(value)))
        return "%s(%s)" % (self.__class__.__name__, ", ".join(l))

    def __eq__(self, other):
        if not isinstance(other, _ttinfo):
            return NotImplemented

        return (self.offset == other.offset and
                self.delta == other.delta and
                self.isdst == other.isdst and
                self.abbr == other.abbr and
                self.isstd == other.isstd and
                self.isgmt == other.isgmt and
                self.dstoffset == other.dstoffset)

    __hash__ = None

    def __ne__(self, other):
        return not (self == other)

    def __getstate__(self):
        state = {}
        for name in self.__slots__:
            state[name] = getattr(self, name, None)
        return state

    def __setstate__(self, state):
        for name in self.__slots__:
            if name in state:
                setattr(self, name, state[name])


class _tzfile(object):
    """
    Lightweight class for holding the relevant transition and time zone
    information read from binary tzfiles.
    """
    attrs = ['trans_list', 'trans_list_utc', 'trans_idx', 'ttinfo_list',
             'ttinfo_std', 'ttinfo_dst', 'ttinfo_before', 'ttinfo_first']

    def __init__(self, **kwargs):
        for attr in self.attrs:
            setattr(self, attr, kwargs.get(attr, None))


class tzfile(_tzinfo):
    """
    This is a ``tzinfo`` subclass thant allows one to use the ``tzfile(5)``
    format timezone files to extract current and historical zone information.

    :param fileobj:
        This can be an opened file stream or a file name that the time zone
        information can be read from.

    :param filename:
        This is an optional parameter specifying the source of the time zone
        information in the event that ``fileobj`` is a file object. If omitted
        and ``fileobj`` is a file stream, this parameter will be set either to
        ``fileobj``'s ``name`` attribute or to ``repr(fileobj)``.

    See `Sources for Time Zone and Daylight Saving Time Data
    <http://www.twinsun.com/tz/tz-link.htm>`_ for more information. Time zone
    files can be compiled from the `IANA Time Zone database files
    <https://www.iana.org/time-zones>`_ with the `zic time zone compiler
    <https://www.freebsd.org/cgi/man.cgi?query=zic&sektion=8>`_
    """

    def __init__(self, fileobj, filename=None):
        super(tzfile, self).__init__()

        file_opened_here = False
        if isinstance(fileobj, string_types):
            self._filename = fileobj
            fileobj = open(fileobj, 'rb')
            file_opened_here = True
        elif filename is not None:
            self._filename = filename
        elif hasattr(fileobj, "name"):
            self._filename = fileobj.name
        else:
            self._filename = repr(fileobj)

        if fileobj is not None:
            if not file_opened_here:
                fileobj = _ContextWrapper(fileobj)

            with fileobj as file_stream:
                tzobj = self._read_tzfile(file_stream)

            self._set_tzdata(tzobj)

    def _set_tzdata(self, tzobj):
        """ Set the time zone data of this object from a _tzfile object """
        # Copy the relevant attributes over as private attributes
        for attr in _tzfile.attrs:
            setattr(self, '_' + attr, getattr(tzobj, attr))

    def _read_tzfile(self, fileobj):
        out = _tzfile()

        # From tzfile(5):
        #
        # The time zone information files used by tzset(3)
        # begin with the magic characters "TZif" to identify
        # them as time zone information files, followed by
        # sixteen bytes reserved for future use, followed by
        # six four-byte values of type long, written in a
        # ``standard'' byte order (the high-order  byte
        # of the value is written first).
        if fileobj.read(4).decode() != "TZif":
            raise ValueError("magic not found")

        fileobj.read(16)

        (
            # The number of UTC/local indicators stored in the file.
            ttisgmtcnt,

            # The number of standard/wall indicators stored in the file.
            ttisstdcnt,

            # The number of leap seconds for which data is
            # stored in the file.
            leapcnt,

            # The number of "transition times" for which data
            # is stored in the file.
            timecnt,

            # The number of "local time types" for which data
            # is stored in the file (must not be zero).
            typecnt,

            # The  number  of  characters  of "time zone
            # abbreviation strings" stored in the file.
            charcnt,

        ) = struct.unpack(">6l", fileobj.read(24))

        # The above header is followed by tzh_timecnt four-byte
        # values  of  type long,  sorted  in ascending order.
        # These values are written in ``standard'' byte order.
        # Each is used as a transition time (as  returned  by
        # time(2)) at which the rules for computing local time
        # change.

        if timecnt:
            out.trans_list_utc = list(struct.unpack(">%dl" % timecnt,
                                                    fileobj.read(timecnt*4)))
        else:
            out.trans_list_utc = []

        # Next come tzh_timecnt one-byte values of type unsigned
        # char; each one tells which of the different types of
        # ``local time'' types described in the file is associated
        # with the same-indexed transition time. These values
        # serve as indices into an array of ttinfo structures that
        # appears next in the file.

        if timecnt:
            out.trans_idx = struct.unpack(">%dB" % timecnt,
                                            fileobj.read(timecnt))
        else:
            out.trans_idx = []

        # Each ttinfo structure is written as a four-byte value
        # for tt_gmtoff  of  type long,  in  a  standard  byte
        # order, followed  by a one-byte value for tt_isdst
        # and a one-byte  value  for  tt_abbrind.   In  each
        # structure, tt_gmtoff  gives  the  number  of
        # seconds to be added to UTC, tt_isdst tells whether
        # tm_isdst should be set by  localtime(3),  and
        # tt_abbrind serves  as an index into the array of
        # time zone abbreviation characters that follow the
        # ttinfo structure(s) in the file.

        ttinfo = []

        for i in range(typecnt):
            ttinfo.append(struct.unpack(">lbb", fileobj.read(6)))

        abbr = fileobj.read(charcnt).decode()

        # Then there are tzh_leapcnt pairs of four-byte
        # values, written in  standard byte  order;  the
        # first  value  of  each pair gives the time (as
        # returned by time(2)) at which a leap second
        # occurs;  the  second  gives the  total  number of
        # leap seconds to be applied after the given time.
        # The pairs of values are sorted in ascending order
        # by time.

        # Not used, for now (but seek for correct file position)
        if leapcnt:
            fileobj.seek(leapcnt * 8, os.SEEK_CUR)

        # Then there are tzh_ttisstdcnt standard/wall
        # indicators, each stored as a one-byte value;
        # they tell whether the transition times associated
        # with local time types were specified as standard
        # time or wall clock time, and are used when
        # a time zone file is used in handling POSIX-style
        # time zone environment variables.

        if ttisstdcnt:
            isstd = struct.unpack(">%db" % ttisstdcnt,
                                  fileobj.read(ttisstdcnt))

        # Finally, there are tzh_ttisgmtcnt UTC/local
        # indicators, each stored as a one-byte value;
        # they tell whether the transition times associated
        # with local time types were specified as UTC or
        # local time, and are used when a time zone file
        # is used in handling POSIX-style time zone envi-
        # ronment variables.

        if ttisgmtcnt:
            isgmt = struct.unpack(">%db" % ttisgmtcnt,
                                  fileobj.read(ttisgmtcnt))

        # Build ttinfo list
        out.ttinfo_list = []
        for i in range(typecnt):
            gmtoff, isdst, abbrind = ttinfo[i]
            # Round to full-minutes if that's not the case. Python's
            # datetime doesn't accept sub-minute timezones. Check
            # http://python.org/sf/1447945 for some information.
            gmtoff = 60 * ((gmtoff + 30) // 60)
            tti = _ttinfo()
            tti.offset = gmtoff
            tti.dstoffset = datetime.timedelta(0)
            tti.delta = datetime.timedelta(seconds=gmtoff)
            tti.isdst = isdst
            tti.abbr = abbr[abbrind:abbr.find('\x00', abbrind)]
            tti.isstd = (ttisstdcnt > i and isstd[i] != 0)
            tti.isgmt = (ttisgmtcnt > i and isgmt[i] != 0)
            out.ttinfo_list.append(tti)

        # Replace ttinfo indexes for ttinfo objects.
        out.trans_idx = [out.ttinfo_list[idx] for idx in out.trans_idx]

        # Set standard, dst, and before ttinfos. before will be
        # used when a given time is before any transitions,
        # and will be set to the first non-dst ttinfo, or to
        # the first dst, if all of them are dst.
        out.ttinfo_std = None
        out.ttinfo_dst = None
        out.ttinfo_before = None
        if out.ttinfo_list:
            if not out.trans_list_utc:
                out.ttinfo_std = out.ttinfo_first = out.ttinfo_list[0]
            else:
                for i in range(timecnt-1, -1, -1):
                    tti = out.trans_idx[i]
                    if not out.ttinfo_std and not tti.isdst:
                        out.ttinfo_std = tti
                    elif not out.ttinfo_dst and tti.isdst:
                        out.ttinfo_dst = tti

                    if out.ttinfo_std and out.ttinfo_dst:
                        break
                else:
                    if out.ttinfo_dst and not out.ttinfo_std:
                        out.ttinfo_std = out.ttinfo_dst

                for tti in out.ttinfo_list:
                    if not tti.isdst:
                        out.ttinfo_before = tti
                        break
                else:
                    out.ttinfo_before = out.ttinfo_list[0]

        # Now fix transition times to become relative to wall time.
        #
        # I'm not sure about this. In my tests, the tz source file
        # is setup to wall time, and in the binary file isstd and
        # isgmt are off, so it should be in wall time. OTOH, it's
        # always in gmt time. Let me know if you have comments
        # about this.
        laststdoffset = None
        out.trans_list = []
        for i, tti in enumerate(out.trans_idx):
            if not tti.isdst:
                offset = tti.offset
                laststdoffset = offset
            else:
                if laststdoffset is not None:
                    # Store the DST offset as well and update it in the list
                    tti.dstoffset = tti.offset - laststdoffset
                    out.trans_idx[i] = tti

                offset = laststdoffset or 0

            out.trans_list.append(out.trans_list_utc[i] + offset)

        # In case we missed any DST offsets on the way in for some reason, make
        # a second pass over the list, looking for the /next/ DST offset.
        laststdoffset = None
        for i in reversed(range(len(out.trans_idx))):
            tti = out.trans_idx[i]
            if tti.isdst:
                if not (tti.dstoffset or laststdoffset is None):
                    tti.dstoffset = tti.offset - laststdoffset
            else:
                laststdoffset = tti.offset

            if not isinstance(tti.dstoffset, datetime.timedelta):
                tti.dstoffset = datetime.timedelta(seconds=tti.dstoffset)

            out.trans_idx[i] = tti

        out.trans_idx = tuple(out.trans_idx)
        out.trans_list = tuple(out.trans_list)
        out.trans_list_utc = tuple(out.trans_list_utc)

        return out

    def _find_last_transition(self, dt, in_utc=False):
        # If there's no list, there are no transitions to find
        if not self._trans_list:
            return None

        timestamp = _datetime_to_timestamp(dt)

        # Find where the timestamp fits in the transition list - if the
        # timestamp is a transition time, it's part of the "after" period.
        trans_list = self._trans_list_utc if in_utc else self._trans_list
        idx = bisect.bisect_right(trans_list, timestamp)

        # We want to know when the previous transition was, so subtract off 1
        return idx - 1        

    def _get_ttinfo(self, idx):
        # For no list or after the last transition, default to _ttinfo_std
        if idx is None or (idx + 1) >= len(self._trans_list):
            return self._ttinfo_std

        # If there is a list and the time is before it, return _ttinfo_before
        if idx < 0:
            return self._ttinfo_before

        return self._trans_idx[idx]

    def _find_ttinfo(self, dt):
        idx = self._resolve_ambiguous_time(dt)

        return self._get_ttinfo(idx)

    def fromutc(self, dt):
        """
        The ``tzfile`` implementation of :py:func:`datetime.tzinfo.fromutc`.

        :param dt:
            A :py:class:`datetime.datetime` object.

        :raises TypeError:
            Raised if ``dt`` is not a :py:class:`datetime.datetime` object.

        :raises ValueError:
            Raised if this is called with a ``dt`` which does not have this
            ``tzinfo`` attached.

        :return:
            Returns a :py:class:`datetime.datetime` object representing the
            wall time in ``self``'s time zone.
        """
        # These isinstance checks are in datetime.tzinfo, so we'll preserve
        # them, even if we don't care about duck typing.
        if not isinstance(dt, datetime.datetime):
            raise TypeError("fromutc() requires a datetime argument")

        if dt.tzinfo is not self:
            raise ValueError("dt.tzinfo is not self")

        # First treat UTC as wall time and get the transition we're in.
        idx = self._find_last_transition(dt, in_utc=True)
        tti = self._get_ttinfo(idx)

        dt_out = dt + datetime.timedelta(seconds=tti.offset)

        fold = self.is_ambiguous(dt_out, idx=idx)

        return enfold(dt_out, fold=int(fold))

    def is_ambiguous(self, dt, idx=None):
        """
        Whether or not the "wall time" of a given datetime is ambiguous in this
        zone.

        :param dt:
            A :py:class:`datetime.datetime`, naive or time zone aware.


        :return:
            Returns ``True`` if ambiguous, ``False`` otherwise.

        .. versionadded:: 2.6.0
        """
        if idx is None:
            idx = self._find_last_transition(dt)

        # Calculate the difference in offsets from current to previous
        timestamp = _datetime_to_timestamp(dt)
        tti = self._get_ttinfo(idx)

        if idx is None or idx <= 0:
            return False

        od = self._get_ttinfo(idx - 1).offset - tti.offset
        tt = self._trans_list[idx]          # Transition time

        return timestamp < tt + od

    def _resolve_ambiguous_time(self, dt):
        idx = self._find_last_transition(dt)

        # If we have no transitions, return the index
        _fold = self._fold(dt)
        if idx is None or idx == 0:
            return idx

        # If it's ambiguous and we're in a fold, shift to a different index.
        idx_offset = int(not _fold and self.is_ambiguous(dt, idx))

        return idx - idx_offset

    def utcoffset(self, dt):
        if dt is None:
            return None

        if not self._ttinfo_std:
            return ZERO

        return self._find_ttinfo(dt).delta

    def dst(self, dt):
        if dt is None:
            return None

        if not self._ttinfo_dst:
            return ZERO

        tti = self._find_ttinfo(dt)

        if not tti.isdst:
            return ZERO

        # The documentation says that utcoffset()-dst() must
        # be constant for every dt.
        return tti.dstoffset

    @tzname_in_python2
    def tzname(self, dt):
        if not self._ttinfo_std or dt is None:
            return None
        return self._find_ttinfo(dt).abbr

    def __eq__(self, other):
        if not isinstance(other, tzfile):
            return NotImplemented
        return (self._trans_list == other._trans_list and
                self._trans_idx == other._trans_idx and
                self._ttinfo_list == other._ttinfo_list)

    __hash__ = None

    def __ne__(self, other):
        return not (self == other)

    def __repr__(self):
        return "%s(%s)" % (self.__class__.__name__, repr(self._filename))

    def __reduce__(self):
        return self.__reduce_ex__(None)

    def __reduce_ex__(self, protocol):
        return (self.__class__, (None, self._filename), self.__dict__)


class tzrange(tzrangebase):
    """
    The ``tzrange`` object is a time zone specified by a set of offsets and
    abbreviations, equivalent to the way the ``TZ`` variable can be specified
    in POSIX-like systems, but using Python delta objects to specify DST
    start, end and offsets.

    :param stdabbr:
        The abbreviation for standard time (e.g. ``'EST'``).

    :param stdoffset:
        An integer or :class:`datetime.timedelta` object or equivalent
        specifying the base offset from UTC.

        If unspecified, +00:00 is used.

    :param dstabbr:
        The abbreviation for DST / "Summer" time (e.g. ``'EDT'``).

        If specified, with no other DST information, DST is assumed to occur
        and the default behavior or ``dstoffset``, ``start`` and ``end`` is
        used. If unspecified and no other DST information is specified, it
        is assumed that this zone has no DST.

        If this is unspecified and other DST information is *is* specified,
        DST occurs in the zone but the time zone abbreviation is left
        unchanged.

    :param dstoffset:
        A an integer or :class:`datetime.timedelta` object or equivalent
        specifying the UTC offset during DST. If unspecified and any other DST
        information is specified, it is assumed to be the STD offset +1 hour.

    :param start:
        A :class:`relativedelta.relativedelta` object or equivalent specifying
        the time and time of year that daylight savings time starts. To specify,
        for example, that DST starts at 2AM on the 2nd Sunday in March, pass:

            ``relativedelta(hours=2, month=3, day=1, weekday=SU(+2))``

        If unspecified and any other DST information is specified, the default
        value is 2 AM on the first Sunday in April.

    :param end:
        A :class:`relativedelta.relativedelta` object or equivalent representing
        the time and time of year that daylight savings time ends, with the
        same specification method as in ``start``. One note is that this should
        point to the first time in the *standard* zone, so if a transition
        occurs at 2AM in the DST zone and the clocks are set back 1 hour to 1AM,
        set the `hours` parameter to +1.


    **Examples:**

    .. testsetup:: tzrange

        from dateutil.tz import tzrange, tzstr

    .. doctest:: tzrange

        >>> tzstr('EST5EDT') == tzrange("EST", -18000, "EDT")
        True

        >>> from dateutil.relativedelta import *
        >>> range1 = tzrange("EST", -18000, "EDT")
        >>> range2 = tzrange("EST", -18000, "EDT", -14400,
        ...                  relativedelta(hours=+2, month=4, day=1,
        ...                                weekday=SU(+1)),
        ...                  relativedelta(hours=+1, month=10, day=31,
        ...                                weekday=SU(-1)))
        >>> tzstr('EST5EDT') == range1 == range2
        True

    """
    def __init__(self, stdabbr, stdoffset=None,
                 dstabbr=None, dstoffset=None,
                 start=None, end=None):

        global relativedelta
        from dateutil import relativedelta

        self._std_abbr = stdabbr
        self._dst_abbr = dstabbr

        try:
            stdoffset = _total_seconds(stdoffset)
        except (TypeError, AttributeError):
            pass

        try:
            dstoffset = _total_seconds(dstoffset)
        except (TypeError, AttributeError):
            pass

        if stdoffset is not None:
            self._std_offset = datetime.timedelta(seconds=stdoffset)
        else:
            self._std_offset = ZERO

        if dstoffset is not None:
            self._dst_offset = datetime.timedelta(seconds=dstoffset)
        elif dstabbr and stdoffset is not None:
            self._dst_offset = self._std_offset + datetime.timedelta(hours=+1)
        else:
            self._dst_offset = ZERO

        if dstabbr and start is None:
            self._start_delta = relativedelta.relativedelta(
                hours=+2, month=4, day=1, weekday=relativedelta.SU(+1))
        else:
            self._start_delta = start

        if dstabbr and end is None:
            self._end_delta = relativedelta.relativedelta(
                hours=+1, month=10, day=31, weekday=relativedelta.SU(-1))
        else:
            self._end_delta = end

        self._dst_base_offset_ = self._dst_offset - self._std_offset
        self.hasdst = bool(self._start_delta)

    def transitions(self, year):
        """
        For a given year, get the DST on and off transition times, expressed
        always on the standard time side. For zones with no transitions, this
        function returns ``None``.

        :param year:
            The year whose transitions you would like to query.

        :return:
            Returns a :class:`tuple` of :class:`datetime.datetime` objects,
            ``(dston, dstoff)`` for zones with an annual DST transition, or
            ``None`` for fixed offset zones.
        """
        if not self.hasdst:
            return None

        base_year = datetime.datetime(year, 1, 1)

        start = base_year + self._start_delta
        end = base_year + self._end_delta

        return (start, end)

    def __eq__(self, other):
        if not isinstance(other, tzrange):
            return NotImplemented

        return (self._std_abbr == other._std_abbr and
                self._dst_abbr == other._dst_abbr and
                self._std_offset == other._std_offset and
                self._dst_offset == other._dst_offset and
                self._start_delta == other._start_delta and
                self._end_delta == other._end_delta)

    @property
    def _dst_base_offset(self):
        return self._dst_base_offset_


class tzstr(tzrange):
    """
    ``tzstr`` objects are time zone objects specified by a time-zone string as
    it would be passed to a ``TZ`` variable on POSIX-style systems (see
    the `GNU C Library: TZ Variable`_ for more details).

    There is one notable exception, which is that POSIX-style time zones use an
    inverted offset format, so normally ``GMT+3`` would be parsed as an offset
    3 hours *behind* GMT. The ``tzstr`` time zone object will parse this as an
    offset 3 hours *ahead* of GMT. If you would like to maintain the POSIX
    behavior, pass a ``True`` value to ``posix_offset``.

    The :class:`tzrange` object provides the same functionality, but is
    specified using :class:`relativedelta.relativedelta` objects. rather than
    strings.

    :param s:
        A time zone string in ``TZ`` variable format. This can be a
        :class:`bytes` (2.x: :class:`str`), :class:`str` (2.x: :class:`unicode`)
        or a stream emitting unicode characters (e.g. :class:`StringIO`).

    :param posix_offset:
        Optional. If set to ``True``, interpret strings such as ``GMT+3`` or
        ``UTC+3`` as being 3 hours *behind* UTC rather than ahead, per the
        POSIX standard.

    .. _`GNU C Library: TZ Variable`:
        https://www.gnu.org/software/libc/manual/html_node/TZ-Variable.html
    """
    def __init__(self, s, posix_offset=False):
        global parser
        from dateutil import parser

        self._s = s

        res = parser._parsetz(s)
        if res is None:
            raise ValueError("unknown string format")

        # Here we break the compatibility with the TZ variable handling.
        # GMT-3 actually *means* the timezone -3.
        if res.stdabbr in ("GMT", "UTC") and not posix_offset:
            res.stdoffset *= -1

        # We must initialize it first, since _delta() needs
        # _std_offset and _dst_offset set. Use False in start/end
        # to avoid building it two times.
        tzrange.__init__(self, res.stdabbr, res.stdoffset,
                         res.dstabbr, res.dstoffset,
                         start=False, end=False)

        if not res.dstabbr:
            self._start_delta = None
            self._end_delta = None
        else:
            self._start_delta = self._delta(res.start)
            if self._start_delta:
                self._end_delta = self._delta(res.end, isend=1)

        self.hasdst = bool(self._start_delta)

    def _delta(self, x, isend=0):
        from dateutil import relativedelta
        kwargs = {}
        if x.month is not None:
            kwargs["month"] = x.month
            if x.weekday is not None:
                kwargs["weekday"] = relativedelta.weekday(x.weekday, x.week)
                if x.week > 0:
                    kwargs["day"] = 1
                else:
                    kwargs["day"] = 31
            elif x.day:
                kwargs["day"] = x.day
        elif x.yday is not None:
            kwargs["yearday"] = x.yday
        elif x.jyday is not None:
            kwargs["nlyearday"] = x.jyday
        if not kwargs:
            # Default is to start on first sunday of april, and end
            # on last sunday of october.
            if not isend:
                kwargs["month"] = 4
                kwargs["day"] = 1
                kwargs["weekday"] = relativedelta.SU(+1)
            else:
                kwargs["month"] = 10
                kwargs["day"] = 31
                kwargs["weekday"] = relativedelta.SU(-1)
        if x.time is not None:
            kwargs["seconds"] = x.time
        else:
            # Default is 2AM.
            kwargs["seconds"] = 7200
        if isend:
            # Convert to standard time, to follow the documented way
            # of working with the extra hour. See the documentation
            # of the tzinfo class.
            delta = self._dst_offset - self._std_offset
            kwargs["seconds"] -= delta.seconds + delta.days * 86400
        return relativedelta.relativedelta(**kwargs)

    def __repr__(self):
        return "%s(%s)" % (self.__class__.__name__, repr(self._s))


class _tzicalvtzcomp(object):
    def __init__(self, tzoffsetfrom, tzoffsetto, isdst,
                 tzname=None, rrule=None):
        self.tzoffsetfrom = datetime.timedelta(seconds=tzoffsetfrom)
        self.tzoffsetto = datetime.timedelta(seconds=tzoffsetto)
        self.tzoffsetdiff = self.tzoffsetto - self.tzoffsetfrom
        self.isdst = isdst
        self.tzname = tzname
        self.rrule = rrule


class _tzicalvtz(_tzinfo):
    def __init__(self, tzid, comps=[]):
        super(_tzicalvtz, self).__init__()

        self._tzid = tzid
        self._comps = comps
        self._cachedate = []
        self._cachecomp = []

    def _find_comp(self, dt):
        if len(self._comps) == 1:
            return self._comps[0]

        dt = dt.replace(tzinfo=None)

        try:
            return self._cachecomp[self._cachedate.index((dt, self._fold(dt)))]
        except ValueError:
            pass

        lastcompdt = None
        lastcomp = None

        for comp in self._comps:
            compdt = self._find_compdt(comp, dt)

            if compdt and (not lastcompdt or lastcompdt < compdt):
                lastcompdt = compdt
                lastcomp = comp

        if not lastcomp:
            # RFC says nothing about what to do when a given
            # time is before the first onset date. We'll look for the
            # first standard component, or the first component, if
            # none is found.
            for comp in self._comps:
                if not comp.isdst:
                    lastcomp = comp
                    break
            else:
                lastcomp = comp[0]

        self._cachedate.insert(0, (dt, self._fold(dt)))
        self._cachecomp.insert(0, lastcomp)

        if len(self._cachedate) > 10:
            self._cachedate.pop()
            self._cachecomp.pop()

        return lastcomp

    def _find_compdt(self, comp, dt):
        if comp.tzoffsetdiff < ZERO and self._fold(dt):
            dt -= comp.tzoffsetdiff

        compdt = comp.rrule.before(dt, inc=True)

        return compdt

    def utcoffset(self, dt):
        if dt is None:
            return None

        return self._find_comp(dt).tzoffsetto

    def dst(self, dt):
        comp = self._find_comp(dt)
        if comp.isdst:
            return comp.tzoffsetdiff
        else:
            return ZERO

    @tzname_in_python2
    def tzname(self, dt):
        return self._find_comp(dt).tzname

    def __repr__(self):
        return "<tzicalvtz %s>" % repr(self._tzid)

    __reduce__ = object.__reduce__


class tzical(object):
    """
    This object is designed to parse an iCalendar-style ``VTIMEZONE`` structure
    as set out in `RFC 2445`_ Section 4.6.5 into one or more `tzinfo` objects.

    :param `fileobj`:
        A file or stream in iCalendar format, which should be UTF-8 encoded
        with CRLF endings.

    .. _`RFC 2445`: https://www.ietf.org/rfc/rfc2445.txt
    """
    def __init__(self, fileobj):
        global rrule
        from dateutil import rrule

        if isinstance(fileobj, string_types):
            self._s = fileobj
            # ical should be encoded in UTF-8 with CRLF
            fileobj = open(fileobj, 'r')
        else:
            self._s = getattr(fileobj, 'name', repr(fileobj))
            fileobj = _ContextWrapper(fileobj)

        self._vtz = {}

        with fileobj as fobj:
            self._parse_rfc(fobj.read())

    def keys(self):
        """
        Retrieves the available time zones as a list.
        """
        return list(self._vtz.keys())

    def get(self, tzid=None):
        """
        Retrieve a :py:class:`datetime.tzinfo` object by its ``tzid``.

        :param tzid:
            If there is exactly one time zone available, omitting ``tzid``
            or passing :py:const:`None` value returns it. Otherwise a valid
            key (which can be retrieved from :func:`keys`) is required.

        :raises ValueError:
            Raised if ``tzid`` is not specified but there are either more
            or fewer than 1 zone defined.

        :returns:
            Returns either a :py:class:`datetime.tzinfo` object representing
            the relevant time zone or :py:const:`None` if the ``tzid`` was
            not found.
        """
        if tzid is None:
            if len(self._vtz) == 0:
                raise ValueError("no timezones defined")
            elif len(self._vtz) > 1:
                raise ValueError("more than one timezone available")
            tzid = next(iter(self._vtz))

        return self._vtz.get(tzid)

    def _parse_offset(self, s):
        s = s.strip()
        if not s:
            raise ValueError("empty offset")
        if s[0] in ('+', '-'):
            signal = (-1, +1)[s[0] == '+']
            s = s[1:]
        else:
            signal = +1
        if len(s) == 4:
            return (int(s[:2]) * 3600 + int(s[2:]) * 60) * signal
        elif len(s) == 6:
            return (int(s[:2]) * 3600 + int(s[2:4]) * 60 + int(s[4:])) * signal
        else:
            raise ValueError("invalid offset: " + s)

    def _parse_rfc(self, s):
        lines = s.splitlines()
        if not lines:
            raise ValueError("empty string")

        # Unfold
        i = 0
        while i < len(lines):
            line = lines[i].rstrip()
            if not line:
                del lines[i]
            elif i > 0 and line[0] == " ":
                lines[i-1] += line[1:]
                del lines[i]
            else:
                i += 1

        tzid = None
        comps = []
        invtz = False
        comptype = None
        for line in lines:
            if not line:
                continue
            name, value = line.split(':', 1)
            parms = name.split(';')
            if not parms:
                raise ValueError("empty property name")
            name = parms[0].upper()
            parms = parms[1:]
            if invtz:
                if name == "BEGIN":
                    if value in ("STANDARD", "DAYLIGHT"):
                        # Process component
                        pass
                    else:
                        raise ValueError("unknown component: "+value)
                    comptype = value
                    founddtstart = False
                    tzoffsetfrom = None
                    tzoffsetto = None
                    rrulelines = []
                    tzname = None
                elif name == "END":
                    if value == "VTIMEZONE":
                        if comptype:
                            raise ValueError("component not closed: "+comptype)
                        if not tzid:
                            raise ValueError("mandatory TZID not found")
                        if not comps:
                            raise ValueError(
                                "at least one component is needed")
                        # Process vtimezone
                        self._vtz[tzid] = _tzicalvtz(tzid, comps)
                        invtz = False
                    elif value == comptype:
                        if not founddtstart:
                            raise ValueError("mandatory DTSTART not found")
                        if tzoffsetfrom is None:
                            raise ValueError(
                                "mandatory TZOFFSETFROM not found")
                        if tzoffsetto is None:
                            raise ValueError(
                                "mandatory TZOFFSETFROM not found")
                        # Process component
                        rr = None
                        if rrulelines:
                            rr = rrule.rrulestr("\n".join(rrulelines),
                                                compatible=True,
                                                ignoretz=True,
                                                cache=True)
                        comp = _tzicalvtzcomp(tzoffsetfrom, tzoffsetto,
                                              (comptype == "DAYLIGHT"),
                                              tzname, rr)
                        comps.append(comp)
                        comptype = None
                    else:
                        raise ValueError("invalid component end: "+value)
                elif comptype:
                    if name == "DTSTART":
                        rrulelines.append(line)
                        founddtstart = True
                    elif name in ("RRULE", "RDATE", "EXRULE", "EXDATE"):
                        rrulelines.append(line)
                    elif name == "TZOFFSETFROM":
                        if parms:
                            raise ValueError(
                                "unsupported %s parm: %s " % (name, parms[0]))
                        tzoffsetfrom = self._parse_offset(value)
                    elif name == "TZOFFSETTO":
                        if parms:
                            raise ValueError(
                                "unsupported TZOFFSETTO parm: "+parms[0])
                        tzoffsetto = self._parse_offset(value)
                    elif name == "TZNAME":
                        if parms:
                            raise ValueError(
                                "unsupported TZNAME parm: "+parms[0])
                        tzname = value
                    elif name == "COMMENT":
                        pass
                    else:
                        raise ValueError("unsupported property: "+name)
                else:
                    if name == "TZID":
                        if parms:
                            raise ValueError(
                                "unsupported TZID parm: "+parms[0])
                        tzid = value
                    elif name in ("TZURL", "LAST-MODIFIED", "COMMENT"):
                        pass
                    else:
                        raise ValueError("unsupported property: "+name)
            elif name == "BEGIN" and value == "VTIMEZONE":
                tzid = None
                comps = []
                invtz = True

    def __repr__(self):
        return "%s(%s)" % (self.__class__.__name__, repr(self._s))


if sys.platform != "win32":
    TZFILES = ["/etc/localtime", "localtime"]
    TZPATHS = ["/usr/share/zoneinfo",
               "/usr/lib/zoneinfo",
               "/usr/share/lib/zoneinfo",
               "/etc/zoneinfo"]
else:
    TZFILES = []
    TZPATHS = []


def gettz(name=None):
    tz = None
    if not name:
        try:
            name = os.environ["TZ"]
        except KeyError:
            pass
    if name is None or name == ":":
        for filepath in TZFILES:
            if not os.path.isabs(filepath):
                filename = filepath
                for path in TZPATHS:
                    filepath = os.path.join(path, filename)
                    if os.path.isfile(filepath):
                        break
                else:
                    continue
            if os.path.isfile(filepath):
                try:
                    tz = tzfile(filepath)
                    break
                except (IOError, OSError, ValueError):
                    pass
        else:
            tz = tzlocal()
    else:
        if name.startswith(":"):
            name = name[:-1]
        if os.path.isabs(name):
            if os.path.isfile(name):
                tz = tzfile(name)
            else:
                tz = None
        else:
            for path in TZPATHS:
                filepath = os.path.join(path, name)
                if not os.path.isfile(filepath):
                    filepath = filepath.replace(' ', '_')
                    if not os.path.isfile(filepath):
                        continue
                try:
                    tz = tzfile(filepath)
                    break
                except (IOError, OSError, ValueError):
                    pass
            else:
                tz = None
                if tzwin is not None:
                    try:
                        tz = tzwin(name)
                    except WindowsError:
                        tz = None

                if not tz:
                    from dateutil.zoneinfo import get_zonefile_instance
                    tz = get_zonefile_instance().get(name)

                if not tz:
                    for c in name:
                        # name must have at least one offset to be a tzstr
                        if c in "0123456789":
                            try:
                                tz = tzstr(name)
                            except ValueError:
                                pass
                            break
                    else:
                        if name in ("GMT", "UTC"):
                            tz = tzutc()
                        elif name in time.tzname:
                            tz = tzlocal()
    return tz


def datetime_exists(dt, tz=None):
    """
    Given a datetime and a time zone, determine whether or not a given datetime
    would fall in a gap.

    :param dt:
        A :class:`datetime.datetime` (whose time zone will be ignored if ``tz``
        is provided.)

    :param tz:
        A :class:`datetime.tzinfo` with support for the ``fold`` attribute. If
        ``None`` or not provided, the datetime's own time zone will be used.

    :return:
        Returns a boolean value whether or not the "wall time" exists in ``tz``.
    """
    if tz is None:
        if dt.tzinfo is None:
            raise ValueError('Datetime is naive and no time zone provided.')
        tz = dt.tzinfo

    dt = dt.replace(tzinfo=None)

    # This is essentially a test of whether or not the datetime can survive
    # a round trip to UTC.
    dt_rt = dt.replace(tzinfo=tz).astimezone(tzutc()).astimezone(tz)
    dt_rt = dt_rt.replace(tzinfo=None)

    return dt == dt_rt


def datetime_ambiguous(dt, tz=None):
    """
    Given a datetime and a time zone, determine whether or not a given datetime
    is ambiguous (i.e if there are two times differentiated only by their DST
    status).

    :param dt:
        A :class:`datetime.datetime` (whose time zone will be ignored if ``tz``
        is provided.)

    :param tz:
        A :class:`datetime.tzinfo` with support for the ``fold`` attribute. If
        ``None`` or not provided, the datetime's own time zone will be used.

    :return:
        Returns a boolean value whether or not the "wall time" is ambiguous in
        ``tz``.

    .. versionadded:: 2.6.0
    """
    if tz is None:
        if dt.tzinfo is None:
            raise ValueError('Datetime is naive and no time zone provided.')

        tz = dt.tzinfo

    # If a time zone defines its own "is_ambiguous" function, we'll use that.
    is_ambiguous_fn = getattr(tz, 'is_ambiguous', None)
    if is_ambiguous_fn is not None:
        try:
            return tz.is_ambiguous(dt)
        except:
            pass

    # If it doesn't come out and tell us it's ambiguous, we'll just check if
    # the fold attribute has any effect on this particular date and time.
    dt = dt.replace(tzinfo=tz)
    wall_0 = enfold(dt, fold=0)
    wall_1 = enfold(dt, fold=1)

    same_offset = wall_0.utcoffset() == wall_1.utcoffset()
    same_dst = wall_0.dst() == wall_1.dst()

    return not (same_offset and same_dst)


def _datetime_to_timestamp(dt):
    """
    Convert a :class:`datetime.datetime` object to an epoch timestamp in seconds
    since January 1, 1970, ignoring the time zone.
    """
    return _total_seconds((dt.replace(tzinfo=None) - EPOCH))


class _ContextWrapper(object):
    """
    Class for wrapping contexts so that they are passed through in a
    with statement.
    """
    def __init__(self, context):
        self.context = context

    def __enter__(self):
        return self.context

    def __exit__(*args, **kwargs):
        pass

# vim:ts=4:sw=4:et



import json
from logging import Logger

from model.error_type import ExternalServicesErrorsType
from interactor.route.menu.get_ruptured_products.dto import (
    GetRupturedProductsResponse,
    GetRupturedProductsRequest
)
from model.exception import ExternalServiceErrors
from validation_util import validate_api_key
from messagebus import MessageBusException

from model.config import Config
from datetime import timedelta, datetime

from model.repository import PosRepositoryModel


class GetRupturedProducts(object):

    def __init__(self, logger, config, pos_repository, cache_menu_ruptured_items):
        # type: (Logger, Config, PosRepositoryModel, dict) -> None

        self.logger = logger
        self.config = config
        self.cache_menu_ruptured_items = cache_menu_ruptured_items
        self.pos_repository = pos_repository

    def execute(self, request):
        # type: (GetRupturedProductsRequest) -> dict

        validate_api_key(request, self.config.api_key_mapping)
        has_cached_data = self.cache_menu_ruptured_items.get("last_request_date")
        if has_cached_data:
            should_renew_cache = (datetime.today() - self.cache_menu_ruptured_items.get("last_request_date")) > \
                                 timedelta(seconds=self.config.ruptured_items_cache_time)
            if should_renew_cache:
                self.create_cache_data()
        else:
            self.create_cache_data()

        return self.cache_menu_ruptured_items.get("data")

    def create_cache_data(self):
        try:
            rupture_items = self.pos_repository.get_rupture_products()
        except MessageBusException:
            raise ExternalServiceErrors(ExternalServicesErrorsType.ServiceUnavailable)
        self.cache_menu_ruptured_items["last_request_date"] = datetime.today()
        response = GetRupturedProductsResponse(json.loads(rupture_items))
        self.cache_menu_ruptured_items["data"] = response.ruptured_products



import json

from mbcontextmessagehandler import MbContextMessageBus
from mwhelper import BaseRepository
from typing import Dict

from persistence import Connection

from model.services import ProductInfoServiceModel


class ProductInfoService(ProductInfoServiceModel, BaseRepository):

    def __init__(
        self,
        message_bus     # type: MbContextMessageBus
    ):
        # type: (...) -> None

        super(ProductInfoService, self).__init__(message_bus.mbcontext)
        self.message_bus = message_bus

    def get_product_info(
        self,
        language    # type: str
    ):
        # type: (...) -> str

        def inner_func(
            conn    # type: Connection
        ):
            # type: (...) -> str

            product_description = self._get_all_products(conn=conn)
            self._fill_description(
                descriptions=product_description,
                language=language,
                conn=conn
            )
            custom_param_dict = self._get_custom_params(conn=conn)
            self._fill_description_custom_params(
                descriptions=product_description,
                custom_params=custom_param_dict
            )
            return product_description

        product_description_json = self.execute_with_connection(
            method=inner_func,
            service_name="Persistence"
        )

        return product_description_json

    @staticmethod
    def _get_all_products(
        conn    # type: Connection
    ):
        # type: (...) -> Dict

        query = """SELECT ProductCode, ProductName FROM Product"""
        cursor = conn.select(sql=query)

        products = {}
        for row in cursor:
            product_id = str(row.get_entry('ProductCode'))
            name = row.get_entry('ProductName')
            products[product_id] = {
                "nutritionalInfo": "",
                "allergens": "",
                "description": "",
                "briefDescription": "",
                "displayName": name
            }

        return products

    @staticmethod
    def _fill_description(
        descriptions,   # type: Dict
        language,       # type: str
        conn            # type: Connection
    ):
        # type: (...) -> None

        query = """
                    SELECT
                        PD.ProductCode AS ProductCode,
                        PD.ProductDescription AS ProductDescription,
                        DES.Description AS Description
                    FROM ProductDescriptions PD
                    JOIN Descriptions DES ON DES.DescrId=PD.DescrId
                    WHERE DES.Language = "{}"
                """.format(language)

        cursor = conn.select(sql=query)

        for row in cursor:
            product_id = str(row.get_entry('ProductCode'))
            description_name = row.get_entry('Description')
            description_content = row.get_entry('ProductDescription')
            if description_name == "nutritionalInfo":
                description_content = json.loads(description_content)

            descriptions[product_id].update({description_name: description_content})

    @staticmethod
    def _get_custom_params(
        conn    # type: Connection
    ):
        # type: (...) -> Dict

        query = """
            SELECT 
                ProductCode, 
                CustomParamValue AS ProductDescription,
                CustomParamId AS Description
            FROM ProductCustomParams 
            WHERE CustomParamId in ('PublicName', 'ProductBriefDescription')
        """

        cursor = conn.select(query)

        custom_param_dict = {}
        for row in cursor:
            product_id = row.get_entry('ProductCode')
            description_type = row.get_entry('Description')
            description_value = row.get_entry('ProductDescription')
            if product_id not in custom_param_dict.keys():
                custom_param_dict[product_id] = {
                    "PublicName": "",
                    "ProductBriefDescription": ""
                }

            custom_param_dict[product_id].update({description_type: description_value})

        return custom_param_dict

    @staticmethod
    def _fill_description_custom_params(
        descriptions,   # type: Dict
        custom_params   # type: Dict
    ):
        # type: (...) -> None

        for product_id in custom_params:
            if product_id not in descriptions.keys():
                continue

            brief_description = custom_params[product_id]["ProductBriefDescription"]
            descriptions[product_id]["briefDescription"] = brief_description

            display_name = custom_params[product_id]["PublicName"]
            if display_name:
                descriptions[product_id]["displayName"] = display_name



# -*- coding: utf-8 -*-

import sys

from xml.etree import cElementTree as eTree
from nfcecanceler import NfceCancelerResponse, NfceCancelerResponseParserException


class NfceCancelerResponseParser(object):
    NAMESPACE_SOAP = "http://www.w3.org/2003/05/soap-envelope"
    NAMESPACE_NFE = "http://www.portalfiscal.inf.br/nfe"
    NAMESPACE_RET_CANCELER = "http://www.portalfiscal.inf.br/nfe/wsdl/RecepcaoEvento"
    NAMESPACE_RET_CANCELER_4 = "http://www.portalfiscal.inf.br/nfe/wsdl/NFeRecepcaoEvento4"

    def parse_response(self, response_str, versao_ws):
        # type: (str, int) -> NfceCancelerResponse
        try:
            response_xml = eTree.XML(response_str)
            if versao_ws in (1, 3):
                namespace_canceler = self.NAMESPACE_RET_CANCELER
                tag_result = "nfeRecepcaoEventoResult"
            else:
                namespace_canceler = self.NAMESPACE_RET_CANCELER_4
                tag_result = "nfeResultMsg"
            c_stat = response_xml.find("{{{0}}}Body/{{{1}}}{2}/{{{3}}}retEnvEvento/{{{3}}}retEvento/{{{3}}}infEvento/{{{3}}}cStat"
                                       .format(NfceCancelerResponseParser.NAMESPACE_SOAP,
                                               namespace_canceler,
                                               tag_result,
                                               NfceCancelerResponseParser.NAMESPACE_NFE)).text

            x_motivo = response_xml.find("{{{0}}}Body/{{{1}}}{2}/{{{3}}}retEnvEvento/{{{3}}}retEvento/{{{3}}}infEvento/{{{3}}}xMotivo"
                                         .format(NfceCancelerResponseParser.NAMESPACE_SOAP,
                                                 namespace_canceler,
                                                 tag_result,
                                                 NfceCancelerResponseParser.NAMESPACE_NFE)).text

            n_prot_node = response_xml.find("{{{0}}}Body/{{{1}}}{2}/{{{3}}}retEnvEvento/{{{3}}}retEvento/{{{3}}}infEvento/{{{3}}}nProt"
                                            .format(NfceCancelerResponseParser.NAMESPACE_SOAP,
                                                    namespace_canceler,
                                                    tag_result,
                                                    NfceCancelerResponseParser.NAMESPACE_NFE))

            n_prot = None
            if n_prot_node is not None:
                n_prot = n_prot_node.text

            return NfceCancelerResponse(c_stat, x_motivo, n_prot)
        except Exception as ex:
            error_message = "Erro parseando XML de response: {0}".format(repr(ex)), sys.exc_info()[2]
            raise NfceCancelerResponseParserException(error_message)



from typing import List, Any

from salechallengereport.model import OperatorSaleItem, ProductSaleItem

from ._SaleChallengeHeaderDto import SaleChallengeHeaderDto


class SaleChallengeGeneratorDto(object):
    def __init__(self, header, items, columns):
        # type: (SaleChallengeHeaderDto, Any[List[OperatorSaleItem], List[ProductSaleItem]], int) -> ()
        self.header = header
        self.columns = columns
        self.items = items



from abc import ABCMeta, abstractmethod

from ._Report import Report


class Formatter(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def format(self, report):
        # type: (Report) -> bytes
        raise NotImplementedError()



data = (
'syae',    # 0x00
'syaeg',    # 0x01
'syaegg',    # 0x02
'syaegs',    # 0x03
'syaen',    # 0x04
'syaenj',    # 0x05
'syaenh',    # 0x06
'syaed',    # 0x07
'syael',    # 0x08
'syaelg',    # 0x09
'syaelm',    # 0x0a
'syaelb',    # 0x0b
'syaels',    # 0x0c
'syaelt',    # 0x0d
'syaelp',    # 0x0e
'syaelh',    # 0x0f
'syaem',    # 0x10
'syaeb',    # 0x11
'syaebs',    # 0x12
'syaes',    # 0x13
'syaess',    # 0x14
'syaeng',    # 0x15
'syaej',    # 0x16
'syaec',    # 0x17
'syaek',    # 0x18
'syaet',    # 0x19
'syaep',    # 0x1a
'syaeh',    # 0x1b
'seo',    # 0x1c
'seog',    # 0x1d
'seogg',    # 0x1e
'seogs',    # 0x1f
'seon',    # 0x20
'seonj',    # 0x21
'seonh',    # 0x22
'seod',    # 0x23
'seol',    # 0x24
'seolg',    # 0x25
'seolm',    # 0x26
'seolb',    # 0x27
'seols',    # 0x28
'seolt',    # 0x29
'seolp',    # 0x2a
'seolh',    # 0x2b
'seom',    # 0x2c
'seob',    # 0x2d
'seobs',    # 0x2e
'seos',    # 0x2f
'seoss',    # 0x30
'seong',    # 0x31
'seoj',    # 0x32
'seoc',    # 0x33
'seok',    # 0x34
'seot',    # 0x35
'seop',    # 0x36
'seoh',    # 0x37
'se',    # 0x38
'seg',    # 0x39
'segg',    # 0x3a
'segs',    # 0x3b
'sen',    # 0x3c
'senj',    # 0x3d
'senh',    # 0x3e
'sed',    # 0x3f
'sel',    # 0x40
'selg',    # 0x41
'selm',    # 0x42
'selb',    # 0x43
'sels',    # 0x44
'selt',    # 0x45
'selp',    # 0x46
'selh',    # 0x47
'sem',    # 0x48
'seb',    # 0x49
'sebs',    # 0x4a
'ses',    # 0x4b
'sess',    # 0x4c
'seng',    # 0x4d
'sej',    # 0x4e
'sec',    # 0x4f
'sek',    # 0x50
'set',    # 0x51
'sep',    # 0x52
'seh',    # 0x53
'syeo',    # 0x54
'syeog',    # 0x55
'syeogg',    # 0x56
'syeogs',    # 0x57
'syeon',    # 0x58
'syeonj',    # 0x59
'syeonh',    # 0x5a
'syeod',    # 0x5b
'syeol',    # 0x5c
'syeolg',    # 0x5d
'syeolm',    # 0x5e
'syeolb',    # 0x5f
'syeols',    # 0x60
'syeolt',    # 0x61
'syeolp',    # 0x62
'syeolh',    # 0x63
'syeom',    # 0x64
'syeob',    # 0x65
'syeobs',    # 0x66
'syeos',    # 0x67
'syeoss',    # 0x68
'syeong',    # 0x69
'syeoj',    # 0x6a
'syeoc',    # 0x6b
'syeok',    # 0x6c
'syeot',    # 0x6d
'syeop',    # 0x6e
'syeoh',    # 0x6f
'sye',    # 0x70
'syeg',    # 0x71
'syegg',    # 0x72
'syegs',    # 0x73
'syen',    # 0x74
'syenj',    # 0x75
'syenh',    # 0x76
'syed',    # 0x77
'syel',    # 0x78
'syelg',    # 0x79
'syelm',    # 0x7a
'syelb',    # 0x7b
'syels',    # 0x7c
'syelt',    # 0x7d
'syelp',    # 0x7e
'syelh',    # 0x7f
'syem',    # 0x80
'syeb',    # 0x81
'syebs',    # 0x82
'syes',    # 0x83
'syess',    # 0x84
'syeng',    # 0x85
'syej',    # 0x86
'syec',    # 0x87
'syek',    # 0x88
'syet',    # 0x89
'syep',    # 0x8a
'syeh',    # 0x8b
'so',    # 0x8c
'sog',    # 0x8d
'sogg',    # 0x8e
'sogs',    # 0x8f
'son',    # 0x90
'sonj',    # 0x91
'sonh',    # 0x92
'sod',    # 0x93
'sol',    # 0x94
'solg',    # 0x95
'solm',    # 0x96
'solb',    # 0x97
'sols',    # 0x98
'solt',    # 0x99
'solp',    # 0x9a
'solh',    # 0x9b
'som',    # 0x9c
'sob',    # 0x9d
'sobs',    # 0x9e
'sos',    # 0x9f
'soss',    # 0xa0
'song',    # 0xa1
'soj',    # 0xa2
'soc',    # 0xa3
'sok',    # 0xa4
'sot',    # 0xa5
'sop',    # 0xa6
'soh',    # 0xa7
'swa',    # 0xa8
'swag',    # 0xa9
'swagg',    # 0xaa
'swags',    # 0xab
'swan',    # 0xac
'swanj',    # 0xad
'swanh',    # 0xae
'swad',    # 0xaf
'swal',    # 0xb0
'swalg',    # 0xb1
'swalm',    # 0xb2
'swalb',    # 0xb3
'swals',    # 0xb4
'swalt',    # 0xb5
'swalp',    # 0xb6
'swalh',    # 0xb7
'swam',    # 0xb8
'swab',    # 0xb9
'swabs',    # 0xba
'swas',    # 0xbb
'swass',    # 0xbc
'swang',    # 0xbd
'swaj',    # 0xbe
'swac',    # 0xbf
'swak',    # 0xc0
'swat',    # 0xc1
'swap',    # 0xc2
'swah',    # 0xc3
'swae',    # 0xc4
'swaeg',    # 0xc5
'swaegg',    # 0xc6
'swaegs',    # 0xc7
'swaen',    # 0xc8
'swaenj',    # 0xc9
'swaenh',    # 0xca
'swaed',    # 0xcb
'swael',    # 0xcc
'swaelg',    # 0xcd
'swaelm',    # 0xce
'swaelb',    # 0xcf
'swaels',    # 0xd0
'swaelt',    # 0xd1
'swaelp',    # 0xd2
'swaelh',    # 0xd3
'swaem',    # 0xd4
'swaeb',    # 0xd5
'swaebs',    # 0xd6
'swaes',    # 0xd7
'swaess',    # 0xd8
'swaeng',    # 0xd9
'swaej',    # 0xda
'swaec',    # 0xdb
'swaek',    # 0xdc
'swaet',    # 0xdd
'swaep',    # 0xde
'swaeh',    # 0xdf
'soe',    # 0xe0
'soeg',    # 0xe1
'soegg',    # 0xe2
'soegs',    # 0xe3
'soen',    # 0xe4
'soenj',    # 0xe5
'soenh',    # 0xe6
'soed',    # 0xe7
'soel',    # 0xe8
'soelg',    # 0xe9
'soelm',    # 0xea
'soelb',    # 0xeb
'soels',    # 0xec
'soelt',    # 0xed
'soelp',    # 0xee
'soelh',    # 0xef
'soem',    # 0xf0
'soeb',    # 0xf1
'soebs',    # 0xf2
'soes',    # 0xf3
'soess',    # 0xf4
'soeng',    # 0xf5
'soej',    # 0xf6
'soec',    # 0xf7
'soek',    # 0xf8
'soet',    # 0xf9
'soep',    # 0xfa
'soeh',    # 0xfb
'syo',    # 0xfc
'syog',    # 0xfd
'syogg',    # 0xfe
'syogs',    # 0xff
)



from abc import ABCMeta, abstractmethod
from ._OperatorTip import OperatorTip


class TipRepository(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def get_tips_per_operator(self, start_date, end_date, operator):
        # type: (datetime, datetime, str) -> List[OperatorTip]
        raise NotImplementedError



from decimal import Decimal
# On Python 2.6 must install lxml since the older xml.etree.ElementTree
# version can not be used to create SVG images.
try:
    import lxml.etree as ET
except ImportError:
    import xml.etree.ElementTree as ET
import qrcode.image.base


class SvgFragmentImage(qrcode.image.base.BaseImage):
    """
    SVG image builder

    Creates a QR-code image as a SVG document fragment.
    """

    _SVG_namespace = "http://www.w3.org/2000/svg"
    kind = "SVG"
    allowed_kinds = ("SVG",)

    def __init__(self, *args, **kwargs):
        ET.register_namespace("svg", self._SVG_namespace)
        super(SvgFragmentImage, self).__init__(*args, **kwargs)
        # Save the unit size, for example the default box_size of 10 is '1mm'.
        self.unit_size = self.units(self.box_size)

    def drawrect(self, row, col):
        self._img.append(self._rect(row, col))

    def units(self, pixels, text=True):
        """
        A box_size of 10 (default) equals 1mm.
        """
        units = Decimal(pixels) / 10
        if not text:
            return units
        return '%smm' % units

    def save(self, stream, kind=None):
        self.check_kind(kind=kind)
        self._write(stream)

    def new_image(self, **kwargs):
        return self._svg()

    def _svg(self, tag=None, version='1.1', **kwargs):
        if tag is None:
            tag = ET.QName(self._SVG_namespace, "svg")
        dimension = self.units(self.pixel_size)
        return ET.Element(
            tag, width=dimension, height=dimension, version=version,
            **kwargs)

    def _rect(self, row, col, tag=None):
        if tag is None:
            tag = ET.QName(self._SVG_namespace, "rect")
        x, y = self.pixel_box(row, col)[0]
        return ET.Element(
            tag, x=self.units(x), y=self.units(y),
            width=self.unit_size, height=self.unit_size)

    def _write(self, stream):
        ET.ElementTree(self._img).write(stream, xml_declaration=False)


class SvgImage(SvgFragmentImage):
    """
    Standalone SVG image builder

    Creates a QR-code image as a standalone SVG document.
    """
    background = None

    def _svg(self, tag='svg', **kwargs):
        svg = super(SvgImage, self)._svg(tag=tag, **kwargs)
        svg.set("xmlns", self._SVG_namespace)
        if self.background:
            svg.append(
                ET.Element(
                    'rect', fill=self.background, x='0', y='0', width='100%',
                    height='100%'))
        return svg

    def _rect(self, row, col):
        return super(SvgImage, self)._rect(row, col, tag="rect")

    def _write(self, stream):
        ET.ElementTree(self._img).write(stream, encoding="UTF-8",
                                        xml_declaration=True)


class SvgPathImage(SvgImage):
    """
    SVG image builder with one single <path> element (removes white spaces
    between individual QR points).
    """

    QR_PATH_STYLE = 'fill:#000000;fill-opacity:1;fill-rule:nonzero;stroke:none'

    def __init__(self, *args, **kwargs):
        self._points = set()
        super(SvgPathImage, self).__init__(*args, **kwargs)

    def _svg(self, viewBox=None, **kwargs):
        if viewBox is None:
            dimension = self.units(self.pixel_size, text=False)
            viewBox = '0 0 %(d)s %(d)s' % {'d': dimension}
        return super(SvgPathImage, self)._svg(viewBox=viewBox, **kwargs)

    def drawrect(self, row, col):
        # (x, y)
        self._points.add((col, row))

    def _generate_subpaths(self):
        """Generates individual QR points as subpaths"""

        rect_size = self.units(self.box_size, text=False)

        for point in self._points:
            x_base = self.units(
                (point[0]+self.border)*self.box_size, text=False)
            y_base = self.units(
                (point[1]+self.border)*self.box_size, text=False)

            yield (
                'M %(x0)s %(y0)s L %(x0)s %(y1)s L %(x1)s %(y1)s L %(x1)s '
                '%(y0)s z' % dict(
                    x0=x_base, y0=y_base,
                    x1=x_base+rect_size, y1=y_base+rect_size,
                ))

    def make_path(self):
        subpaths = self._generate_subpaths()

        return ET.Element(
            ET.QName("path"),
            style=self.QR_PATH_STYLE,
            d=' '.join(subpaths),
            id="qr-path"
        )

    def _write(self, stream):
        self._img.append(self.make_path())
        super(SvgPathImage, self)._write(stream)


class SvgFillImage(SvgImage):
    """
    An SvgImage that fills the background to white.
    """
    background = 'white'


class SvgPathFillImage(SvgPathImage):
    """
    An SvgPathImage that fills the background to white.
    """
    background = 'white'
